/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the mozCellML code.
 *
 * The Initial Developer of the Original Code is
 * The Bioengineering Institute, University of Auckland.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrew Miller <ak.miller@auckland.ac.nz>
 *
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the NPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include "mozCellMLIntegrator.h"
#include <nsArray.h>
#include <nsIEventQueueService.h>
#include <nsIServiceManager.h>
#include "mozCellMLVariable.h"
#include "mozCellMLPointToZinc.h"

NS_IMPL_ISUPPORTS5(mozCellMLIntegrator,
                   mozICellMLIntegrator,
                   nsIInputStreamCallback,
                   nsIOutputStreamCallback,
                   nsIServerSocketListener,
                   nsISupportsWeakReference
                  )

NS_IMPL_ISUPPORTS2(mozCellMLIntegrationTask,
                   mozICellMLIntegrationTask,
                   nsISupportsWeakReference)

mozCellMLIntegrator::mozCellMLIntegrator(
                                         nsIServerSocket* listener,
                                         nsCOMArray<mozCellMLVariable>& idifferentials,
                                         nsCOMArray<mozCellMLVariable>& isimples,
                                         nsCOMArray<mozCellMLVariable>& iconstants,
                                         nsACString& ifilename,
                                         PRUint32 ihidden_higher
                                        )
  : differentials(idifferentials),
    simples(isimples), constants(iconstants), queuedIntegrationTasks(nsnull),
    buf_pos(0), hidden_higher(ihidden_higher), wantcount(1),
    filename(ifilename)
{
  NS_ADDREF(this);
  LEAKTRACE_CREATED(mozCellMLIntegrator)
  nsCOMPtr<nsIServiceManager> servMan;
  nsCOMPtr<nsIEventQueueService> eqs = do_GetService(NS_EVENTQUEUESERVICE_CONTRACTID);
  nsCOMPtr<nsIServerSocketListener> wl = new WeakServerSocketListenerAdaptor(this);
  listener->AsyncListen(wl);
  eqs->GetSpecialEventQueue(nsIEventQueueService::
                            CURRENT_THREAD_EVENT_QUEUE,
                            getter_AddRefs(eventQueue));
  numvars = differentials.Count() + simples.Count();
  variable = new mozICellMLVariable*[numvars];
  PRInt32 i, j;
  for (i = 0; i < simples.Count(); i++)
    variable[i] = simples[i];
  for (j = i; i < differentials.Count() + j; i++)
    variable[i] = differentials[i - j];
  max_packet = (sizeof(IntegratorReplyHeader) +
                (1 + numvars + hidden_higher) * sizeof(double));
  in_buffer = (char*)PR_Malloc(max_packet);
  transport = NULL;
}

mozCellMLIntegrator::~mozCellMLIntegrator()
{
  LEAKTRACE_DELETED(mozCellMLIntegrator)

  if (wantcount > 0)
  {
    wantcount = 1;
    Destroy();
  }
  while (queuedIntegrationTasks.GetSize())
  {
    void* x = queuedIntegrationTasks.Pop();
    nsIWeakReference* wr = NS_STATIC_CAST(nsIWeakReference*, x);
    NS_RELEASE(wr);
  }

  if (variable)
    delete [] variable;

  if (in_buffer)
    PR_Free(in_buffer);

  PR_Delete(filename.get());
}

NS_IMETHODIMP
mozCellMLIntegrator::Want()
{
  wantcount++;
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLIntegrator::Destroy()
{
  struct HelperCommandHeader ich;

  if (--wantcount > 0)
    return NS_OK;

  ich.length = sizeof(ich);
  ich.cmd = STOP_HELPER;
  Enqueue(&ich, sizeof(ich));
  return NS_OK;
}

/**
 * This method adds a task to the task queue.
 */
void
mozCellMLIntegrator::StartTask(mozCellMLIntegrationTask* task)
{
  // Convert to a weak reference...
  nsIWeakReference* weakPtr =
    NS_GetWeakReference(NS_STATIC_CAST(nsISupportsWeakReference*, task));

  // Put the weak reference on the front of the queue.
  queuedIntegrationTasks.PushFront(weakPtr);

  // If there are no tasks running, then launch the task.
  if (activeTask == nsnull)
    LaunchTaskQueue();
}

/**
 * This method checks if there are any tasks on the task queue, and if so, it
 * starts the first task on the queue.
 */
void
mozCellMLIntegrator::LaunchTaskQueue()
{
#ifdef DEBUG_MODE
  printf("LaunchTaskQueue() called.\n");
#endif
  // Wipe out any current active task...
  activeTask = nsnull;

  // Find a new active task...
  do
  {
    nsIWeakReference* wr = (nsIWeakReference*)queuedIntegrationTasks.Pop();
    if (wr == nsnull)
      return;
    nsWeakPtr wp = getter_AddRefs(wr);
    activeTask = do_QueryReferent(wp);

    // Skip cancelled tasks...
    if (activeTask != nsnull)
    {
      mozICellMLIntegrationTask* iat = activeTask;
      mozCellMLIntegrationTask* at = NS_STATIC_CAST(mozCellMLIntegrationTask*, iat);
      if (at->WasTaskCancelled())
        activeTask = nsnull;
    }
  }
  while (activeTask == nsnull);

  // Fetch the start command out of the active task...
  PRUint32 packet_len;
  mozICellMLIntegrationTask* iat = activeTask;
  mozCellMLIntegrationTask* at = NS_STATIC_CAST(mozCellMLIntegrationTask*, iat);
  struct HelperCommandHeader* hdr = at->GetStartCommand(&packet_len);

  // Send the start command...
#ifdef DEBUG_MODE
  printf("Sending start command to the integrator...\n");
#endif
  Enqueue(hdr, packet_len);
}

NS_IMETHODIMP
mozCellMLIntegrator::ProcessPacket(IntegratorReply reply, void* data, PRUint32 length)
{
#ifdef DEBUG_MODE
  printf("Processing a packet %u of length %u\n", reply, length);
#endif

  mozICellMLIntegrationTask* iat = activeTask;
  mozCellMLIntegrationTask* at = NS_STATIC_CAST(mozCellMLIntegrationTask*, iat);

  if (activeTask == nsnull)
    // We got an unexpected packet. This might be because we cancelled the
    // active task and this packet crossed our suspend packet.
    return NS_OK;

  double *ddata = (double*)data;

  switch (reply)
  {
  case INTEGRATION_POINT:
    at->PointAdded(ddata[0], numvars, variable, ddata + 1);
    break;
  case INTEGRATION_DONE:
    at->Done();
    LaunchTaskQueue();
    break;
  case INTEGRATION_FAILED:
    // Data is guaranteed to be null terminated by integrator.
    at->Failed((char*)data);
    LaunchTaskQueue();
    break;
  case INTEGRATION_HALTED:
    at->Halted();
    LaunchTaskQueue();
  }
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLIntegrator::OnSocketAccepted(
                                      nsIServerSocket* servsocket,
                                      nsISocketTransport* newsocket
                                     )
{
#ifdef DEBUG_MODE
  printf("OnSocketAccepted() called.\n");
#endif
  if (transport == NULL)
  {
    transport = newsocket;
    nsCOMPtr<nsIInputStream> input_s;
    transport->OpenInputStream(
                               nsITransport::OPEN_UNBUFFERED, 0, 0,
                               getter_AddRefs(input_s)
                              );
    input = do_QueryInterface(input_s);

    nsCOMPtr<nsIOutputStream> output_s;
    transport->OpenOutputStream(
                                nsITransport::OPEN_UNBUFFERED, 0, 0,
                                getter_AddRefs(output_s)
                               );
    output = do_QueryInterface(output_s);
    
    nsCOMPtr<nsIInputStreamCallback> wica =
      new WeakInputCallbackAdaptor(this);
    input->AsyncWait(wica, 0, 0, eventQueue);
    if (out_buffer.Length() != 0)   
    {
      awaiting_callback = PR_TRUE;
      nsCOMPtr<nsIOutputStreamCallback> woca =
        new WeakOutputCallbackAdaptor(this);
      output->AsyncWait(woca, 0, 0, eventQueue);
    }
    else
      awaiting_callback = PR_FALSE;
    OnOutputStreamReady(output);
  }
  servsocket->Close();
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLIntegrator::OnStopListening(nsIServerSocket* socket, PRUint32 error)
{
  return NS_OK;
}


NS_IMETHODIMP
mozCellMLIntegrator::OnInputStreamReady(nsIAsyncInputStream* ais)
{
#ifdef DEBUG_MODE
  printf("Input stream is ready.\n");
#endif
  // Read as much data as possible...
  PRUint32 actually_read;
  while (PR_TRUE)
  {
    nsresult ret = ais->Read(in_buffer + buf_pos, max_packet - buf_pos,
                             &actually_read);
    if (ret == NS_BASE_STREAM_WOULD_BLOCK)
    {
      nsCOMPtr<nsIInputStreamCallback> wica
        = new WeakInputCallbackAdaptor(this);
      ais->AsyncWait(wica, 0, 0, eventQueue);
      return NS_OK;
    }
    if (ret != NS_OK || actually_read == 0)
      return ret;
    buf_pos += actually_read;
    if (buf_pos < sizeof(IntegratorReplyHeader))
      continue;
    IntegratorReplyHeader* hdr = (IntegratorReplyHeader*)in_buffer;
    PRUint32 hl = hdr->length;
    if (hl < sizeof(IntegratorReplyHeader))
      hl = sizeof(IntegratorReplyHeader);
    if (buf_pos < hl)
      continue;
    ProcessPacket(hdr->reply, (void*)(hdr + 1),
                  hl - sizeof(IntegratorReplyHeader));
    memmove(in_buffer, in_buffer + hl, buf_pos - hdr->length);
    buf_pos -= hl;
  }
}

NS_IMETHODIMP
mozCellMLIntegrator::OnOutputStreamReady(nsIAsyncOutputStream* aos)
{
#ifdef DEBUG_MODE
  printf("Output stream is ready.\n");
#endif
  PRUint32 wrote_length;
  // XXX This is inefficient, we could use a ringbuffer etc...
  while (out_buffer.Length() != 0 &&
         aos->Write(out_buffer.get(), out_buffer.Length(), &wrote_length)
         == NS_OK)
    out_buffer = Substring(out_buffer, wrote_length);

  if (out_buffer.Length() != 0)
  {
    nsCOMPtr<nsIOutputStreamCallback> woca
      = new WeakOutputCallbackAdaptor(this);
    output->AsyncWait(woca, 0, 0, eventQueue);
  }
  else
    awaiting_callback = PR_FALSE;
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLIntegrator::Enqueue(const void* ptr, PRUint32 length)
{
#ifdef DEBUG_MODE
  printf("Enqueue() called\n");
#endif
  out_buffer.Append((char*)ptr, length);
  if (transport != NULL &&
      !awaiting_callback)
  {
    awaiting_callback = PR_TRUE;
    nsCOMPtr<nsIOutputStreamCallback> woca
      = new WeakOutputCallbackAdaptor(this);
    output->AsyncWait(woca, 0, 0, eventQueue);
  }
  return NS_OK;
}

void
mozCellMLIntegrator::ConsiderCancel(mozCellMLIntegrationTask* it)
{
  if (it == activeTask)
  {
    struct HelperCommandHeader ich;
    ich.length = sizeof(ich);
    ich.cmd = SUSPEND_HELPER;
    Enqueue(&ich, sizeof(ich));

    // Don't pull the next task off the queue yet or there could be confusion
    // about which task is currently active.
  }
}

NS_IMETHODIMP
mozCellMLIntegrator::NewTask(mozICellMLIntegrationTask** ret)
{
  *ret = new mozCellMLIntegrationTask(this);
  if (*ret == nsnull)
    return NS_ERROR_OUT_OF_MEMORY;
  NS_ADDREF(*ret);
  return NS_OK;
}

mozCellMLIntegrationTask::mozCellMLIntegrationTask(mozCellMLIntegrator* mci)
  : integrator(mci), hdr(nsnull), wasCancelled(PR_FALSE)
{
}

mozCellMLIntegrationTask::~mozCellMLIntegrationTask()
{
  if (hdr != nsnull)
    PR_Free(hdr);
}

NS_IMETHODIMP
mozCellMLIntegrationTask::Start(
                                mozICellML* model,
                                double startX,
                                double stopX,
                                double resolution,
                                double abs_tolerance,
                                double rel_tolerance,
                                double max_stepsize
                               )
{
  // Start can only be called once per integration task...
  if (hdr != nsnull || integrator == nsnull || wasCancelled)
    return NS_ERROR_FAILURE;

  mozICellMLIntegrator* iintg = integrator;
  mozCellMLIntegrator* intg = NS_STATIC_CAST(mozCellMLIntegrator*, iintg);
  // Compute the length of the packet...
  packet_len = sizeof(struct HelperCommandHeader) +
               sizeof(struct StartIntegrationCommand) +
               (intg->numvars + intg->hidden_higher + intg->constants.Count())
                * sizeof(double);

  if (model == nsnull)
    return NS_ERROR_FAILURE;

  // Set up the packet storage...
  hdr = (struct HelperCommandHeader*)PR_Malloc(packet_len);
  if (hdr == nsnull)
    return NS_ERROR_OUT_OF_MEMORY;
  struct StartIntegrationCommand* cmd = (struct StartIntegrationCommand*)(hdr + 1);
  double* initialValues = (double*)(cmd + 1);

  // Set up the packet and command headers...
  hdr->length = packet_len;
  hdr->cmd = START_HELPER;
  cmd->startX = startX;
  cmd->stopX = stopX;
  cmd->resolution = resolution;
  cmd->abs_tolerance = abs_tolerance;
  cmd->rel_tolerance = rel_tolerance;
  cmd->max_stepsize = max_stepsize;

  // Set up the initial values...
  PRUint32 i;
  for (i = 0; i < intg->numvars; i++)
  {
    nsCOMPtr<mozICellMLVariable> v;
    double val = 0;
    PRUint32 in;
    nsXPIDLString n, c;

    intg->variable[i]->GetName(getter_Copies(n));
    intg->variable[i]->GetComponentName(getter_Copies(c));
    intg->variable[i]->GetImportNumber(&in);
    model->GetVariable(in, c, n, getter_AddRefs(v));
    if (v == NULL)
    {
      PR_Free(hdr);
      return NS_ERROR_FAILURE;
    }
    v->GetInitialValue(&val);
    *initialValues++ = val;
  }
  // XXX the user should be able to specify initial values for hidden higher
  //     variables.
  for (i = 0; i < intg->hidden_higher; i++)
    *initialValues++ = 0;

  PRUint32 l = intg->constants.Count();
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<mozICellMLVariable> v;
    double val = 0;
    nsXPIDLString n, c;
    PRUint32 in;
    intg->constants[i]->GetImportNumber(&in);
    intg->constants[i]->GetComponentName(getter_Copies(c));
    intg->constants[i]->GetName(getter_Copies(n));
    model->GetVariable(in, c, n, getter_AddRefs(v));
    if (v == NULL)
    {
      PR_Free(hdr);
      return NS_ERROR_FAILURE;
    }
    v->GetInitialValue(&val);
    *initialValues++ = val;
  }

  intg->StartTask(this);

  return NS_OK;
}

NS_IMETHODIMP
mozCellMLIntegrationTask::Cancel()
{
  // If we were already cancelled, don't bother...
  if (integrator == nsnull || wasCancelled)
    return NS_OK;

  // Mark this as cancelled.
  wasCancelled = PR_TRUE;
  
  mozICellMLIntegrator* iintg = integrator;
  mozCellMLIntegrator* intg = NS_STATIC_CAST(mozCellMLIntegrator*, iintg);

  // We now inform the integrator. This makes no difference unless we are
  // active, but if we are it has to send a suspend request...
  intg->ConsiderCancel(this);

  return NS_OK;
}

NS_IMETHODIMP
mozCellMLIntegrationTask::RegisterWatcher(mozIPointWatcher* watcher)
{
  if (watcher == NULL)
    return NS_ERROR_FAILURE;
  watchers.AppendObject(NS_GetWeakReference(watcher));
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLIntegrationTask::SendPointsToZinc
(
 mozICellML* aCellML,
 mozIScaleWatcher* scaleWatcher,
 double startx, double stopx, double gridx,
 mozICellMLVariable** variables,
 PRUint32 numvariables,
 CmguiIFunction** functions,
 PRUint32 numfunctions,
 CmguiICommandData* cmddata,
 CmguiITimeSequencePackage* seqPackage,
 CmguiINodeFieldCreator* nfc,
 CmguiINode* pointNode,
 mozIPointToZinc** pointToZinc
)
{
  *pointToZinc = new mozCellMLPointToZinc(aCellML,
                                          scaleWatcher, startx, stopx, gridx,
                                          numvariables, variables, functions,
                                          cmddata, seqPackage, nfc,
                                          pointNode);
  NS_ADDREF(*pointToZinc);
  RegisterWatcher(*pointToZinc);
  return NS_OK;
}

void
mozCellMLIntegrationTask::PointAdded(double point, PRUint32 numvars,
                                     mozICellMLVariable** variables,
                                     double* data)
{
  PRUint32 i, l;
  for (i = 0, l = watchers.Count();
       i < l;
       i++)
  {
    nsCOMPtr<mozIPointWatcher> watcher =
      do_QueryReferent(watchers[i]);
#ifdef DEBUG_MODE
    printf("Checking watcher %u\n", i);
#endif
    if (watcher == nsnull)
    {
#ifdef DEBUG_MODE
      printf("Watcher has been released.\n");
#endif
      continue;
    }

    watcher->PointAdded(point, variables, numvars, data, numvars);
  }
}

void
mozCellMLIntegrationTask::Done()
{  
  PRUint32 i, l;
  for (i = 0, l = watchers.Count();
       i < l;
       i++)
  {
    nsCOMPtr<mozIPointWatcher> watcher =
      do_QueryReferent(watchers[i]);
#ifdef DEBUG_MODE
    printf("Checking watcher %u\n", i);
#endif
    if (watcher == nsnull)
    {
#ifdef DEBUG_MODE
      printf("Watcher has been released.\n");
#endif
      continue;
    }

    watcher->Done();
  }
}

void
mozCellMLIntegrationTask::Halted()
{  
  PRUint32 i, l;
  for (i = 0, l = watchers.Count();
       i < l;
       i++)
  {
    nsCOMPtr<mozIPointWatcher> watcher =
      do_QueryReferent(watchers[i]);
#ifdef DEBUG_MODE
    printf("Checking watcher %u\n", i);
#endif
    if (watcher == nsnull)
    {
#ifdef DEBUG_MODE
      printf("Watcher has been released.\n");
#endif
      continue;
    }

    watcher->Halted();
  }
}

void
mozCellMLIntegrationTask::Failed(const char* errmsg)
{  
  PRUint32 i, l;
  for (i = 0, l = watchers.Count();
       i < l;
       i++)
  {
    nsCOMPtr<mozIPointWatcher> watcher =
      do_QueryReferent(watchers[i]);
#ifdef DEBUG_MODE
    printf("Checking watcher %u\n", i);
#endif
    if (watcher == nsnull)
    {
#ifdef DEBUG_MODE
      printf("Watcher has been released.\n");
#endif
      continue;
    }

    watcher->Failed(errmsg);
  }
}
