/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: NPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Netscape Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the mozCellML code.
 *
 * The Initial Developer of the Original Code is
 * The Bioengineering Institute, University of Auckland.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrew Miller <ak.miller@auckland.ac.nz>
 *
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the NPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "mozCellMLVariable.h"
#include "mozCellMLVariableResolution.h"
#include <prdtoa.h>
#include <nsPrintfCString.h>
#include <nsIArray.h>
#include <nsArray.h>
#include <nsInterfaceHashtable.h>
#include <nsPrintfCString.h>
#include <math.h>

mozCellMLVariable::mozCellMLVariable(nsIDOMElement* var,
                                     mozCellML* mcm)
  : variable(var), AlreadyInIVCall(PR_FALSE),
    seenRuntimeIVAssignment(PR_FALSE), impNo(0), mozcellml(mcm)
{
  LEAKTRACE_CREATED(mozCellMLVariable)
}

mozCellMLVariable::~mozCellMLVariable()
{
  LEAKTRACE_DELETED(mozCellMLVariable)
}

NS_IMPL_ISUPPORTS1(mozCellMLVariable, mozICellMLVariable)

#define T(model, s) (model->Translate(s))
#define COMPLAIN(model,args) \
{\
  char* result = PR_smprintf args; \
  nsCAutoString complainCString(result); \
  nsAutoString complainString; \
  CopyUTF8toUTF16(complainCString, complainString); \
  PR_Free(result); \
  return model->ValidationError(complainString); \
}

nsresult
mozCellMLVariable::SetupUnits(
                              mozCellML* model,
                              nsDataHashtable<nsISupportsHashKey,
                                              PRUint32>&
                                importNumbers,
                              nsInterfaceHashtable<nsStringHashKey,
                                                   mozCellMLUnits>&
                                allUnits
                             )
{
  nsAutoString unitsName;
  nsresult rv = variable->GetAttribute(NS_LITERAL_STRING("units"), unitsName);
  if (rv != NS_OK)
    COMPLAIN(model,(T(model, "variableMustHaveUnits")));

  return FindUnits(model, importNumbers, allUnits, unitsName,
                   getter_AddRefs(units));
}

nsresult
mozCellMLVariable::FindUnits(
                             mozCellML* model,
                             nsDataHashtable<nsISupportsHashKey,
                                             PRUint32>&
                                importNumbers,
                             nsInterfaceHashtable<nsStringHashKey,
                                                  mozCellMLUnits>&
                                allUnits,
                             nsAutoString& unitsName,
                             mozCellMLUnits** foundUnits
                            )
{
  nsCOMPtr<nsIDOMNode> compNode;
  variable->GetParentNode(getter_AddRefs(compNode));
  nsCOMPtr<nsIDOMElement> component(do_QueryInterface(compNode));
  nsCOMPtr<nsIDOMNode> modelNode;
  component->GetParentNode(getter_AddRefs(modelNode));
  nsCOMPtr<nsIDOMElement> modelEl(do_QueryInterface(modelNode));

  PRUint32 impNo;
  importNumbers.Get(modelEl, &impNo);

#ifdef DEBUG_UNITS
  printf("Setting up variable units for import # %u\n", impNo);
#endif

  nsPrintfCString impNoCStr("%u:", impNo);

  nsAutoString impNoStr;
  CopyUTF8toUTF16(impNoCStr, impNoStr);
  
  nsAutoString compName;
  component->GetAttribute(NS_LITERAL_STRING("name"), compName);

  // Attempt 1: In the component...
  nsAutoString findAttempt = impNoStr + compName + NS_LITERAL_STRING(":") +
                             unitsName;
  allUnits.Get(findAttempt, foundUnits);
  if (*foundUnits != nsnull)
    return NS_OK;

  // Attempt 2: In the same model(or imported).
  findAttempt = impNoStr + NS_LITERAL_STRING(":") + unitsName;
  allUnits.Get(findAttempt, foundUnits);
  if (*foundUnits != nsnull)
    return NS_OK;

  // Attempt 3: Built in...
  allUnits.Get(unitsName, foundUnits);
  if (*foundUnits != nsnull)
    return NS_OK;

  nsCAutoString unitsNameC;
  nsCAutoString compNameC;
  CopyUTF16toUTF8(unitsName, unitsNameC);
  CopyUTF16toUTF8(compName, compNameC);
  COMPLAIN(model, (T(model, "cantFindUnits"), unitsNameC.get(),
                   compNameC.get()));
}

void
mozCellMLVariable::Initialise(
                              nsCOMArray<mozCellMLEquation>& allDiffEquations,
                              nsCOMArray<mozCellMLEquation>& allNondiffEquations
                             )
{
  nsresult ret = DetermineType(mozcellml, allDiffEquations,
                               allNondiffEquations);

  if (ret != NS_OK)
  {
    type = 123;
    return;
  }
}

nsresult
mozCellMLVariable::Equals(mozICellMLVariable* compareTo,
                          PRBool* result)
{
  nsXPIDLString myName, theirName;
  compareTo->GetName(getter_Copies(theirName));
  GetName(getter_Copies(myName));
  if (!myName.Equals(theirName))
  {
    *result = PR_FALSE;
    return NS_OK;
  }

  compareTo->GetComponentName(getter_Copies(theirName));
  GetComponentName(getter_Copies(myName));
  if (!myName.Equals(theirName))
  {
    *result = PR_FALSE;
    return NS_OK;
  }

  mozcellml->GetImportNumberPrefix(variable, nsnull, myName);
  // XXX static_cast without checks isn't very nice(e.g. could theoretically
  //     write JS to implement mozICellMLVariable and crash here). But doing
  // a dynamic_cast would kill portability.
  mozcellml->GetImportNumberPrefix(NS_STATIC_CAST(mozCellMLVariable*, compareTo)->variable,
                                   nsnull,
                                   theirName);

  *result = (myName.Equals(theirName));
  return NS_OK;
}

nsresult
mozCellMLVariable::GetDegree(
                             nsIDOMElement* possibleDeg,
                             PRUint32* degr
                            )
{
  nsAutoString localName;
  possibleDeg->GetLocalName(localName);
  if (!localName.Equals(NS_LITERAL_STRING("degree")))
    return NS_ERROR_FAILURE;
  nsCOMPtr<nsIDOMNodeList> subNodes;
  possibleDeg->GetChildNodes(getter_AddRefs(subNodes));
  nsCOMPtr<nsIArray> subText;
  FilterNodeListByType(subNodes, getter_AddRefs(subText),
                         nsIDOMNode::TEXT_NODE);
  PRUint32 numText;
  subText->GetLength(&numText);
  if (numText == 0)
    return NS_ERROR_FAILURE;
  nsCOMPtr<nsISimpleEnumerator> subTextE;
  subText->Enumerate(getter_AddRefs(subTextE));
  nsCOMPtr<nsISupports> txtNodeSupports;
  subTextE->GetNext(getter_AddRefs(txtNodeSupports));
  nsCOMPtr<nsIDOMText> txt(do_QueryInterface(txtNodeSupports));
  nsAutoString valueStr;
  nsCAutoString valueCStr;
  txt->GetData(valueStr);
  valueStr.Trim(" \t\r\n\f");
  CopyUTF16toUTF8(valueStr, valueCStr);
  *degr = strtoul(valueCStr.get(), nsnull, 10);
  return NS_OK;
}

nsresult
mozCellMLVariable::GetCMetaID(nsAString& cmeta)
{
  nsresult rv;
  rv = variable->GetAttributeNS(CMETA_NS_1_0, NS_LITERAL_STRING("id"),
                                cmeta);
  NS_ENSURE_SUCCESS(rv, rv);
  if (cmeta.Length() == 0)
    return NS_ERROR_FAILURE;
  return NS_OK;
}

nsresult
mozCellMLVariable::DetermineType(
                                 mozCellML* mcm,
                                 nsCOMArray<mozCellMLEquation>& allDiffEquations,
                                 nsCOMArray<mozCellMLEquation>& allNondiffEquations
                                )
{
  // Default to simple, unless it has an initial value, in which case
  // default to constant.
  PRBool hasIV = PR_FALSE;
  variable->HasAttribute(NS_LITERAL_STRING("initial_value"), &hasIV);

  if (hasIV)
    type = mozICellMLVariable::CONSTANT;
  else
    type = mozICellMLVariable::SIMPLE;

  // If this is used in any <apply><diff> as a bound, make it bound...
  PRUint32 i;
  for (i = 0; i < (PRUint32)allDiffEquations.Count(); i++)
  {
    mozCellMLEquation::EquationVariableRelation evr;
    nsCOMPtr<mozCellMLVariableResolution> vr;
    allDiffEquations[i]->GetVariableRelation(this, evr, getter_AddRefs(vr));
    if (evr == mozCellMLEquation::BOUND)
    {
      type = mozICellMLVariable::BOUND;
      return NS_OK;
    }
    if (evr == mozCellMLEquation::DIFF_VAR)
    {
      type = mozICellMLVariable::DIFFERENTIAL;
      return NS_OK;
    }
    if (evr == mozCellMLEquation::NOT_FREE)
    {
      // If the variable is used as a bound variable, we exempt it from the
      // requirements that it be defined, by making it a constant.
      type = mozICellMLVariable::CONSTANT;
      return NS_OK;
    }
  }
  for (i = 0; i < (PRUint32)allNondiffEquations.Count(); i++)
  {
    mozCellMLEquation::EquationVariableRelation evr;
    nsCOMPtr<mozCellMLVariableResolution> vr;
    allNondiffEquations[i]->GetVariableRelation(this, evr, getter_AddRefs(vr));
    if (evr == mozCellMLEquation::NOT_FREE)
    {
      // If the variable is used as a bound variable, we exempt it from the
      // requirements that it be defined, by making it a constant.
      type = mozICellMLVariable::CONSTANT;
      return NS_OK;
    }
  }

  return NS_OK;
}

NS_IMETHODIMP
mozCellMLVariable::GetName(PRUnichar** oname)
{
  nsAutoString name;
  nsresult ret =
    variable->GetAttribute(NS_LITERAL_STRING("name"), name);
  *oname = ToNewUnicode(name);
  return ret;
}

NS_IMETHODIMP
mozCellMLVariable::GetComponentName(PRUnichar** oname)
{
  nsCOMPtr<nsIDOMNode> componentNode;
  variable->GetParentNode(getter_AddRefs(componentNode));
  nsCOMPtr<nsIDOMElement> component(do_QueryInterface(componentNode));
  nsAutoString name;
  nsresult ret = component->GetAttribute(NS_LITERAL_STRING("name"), name);
  *oname = ToNewUnicode(name);
  return ret;
}

NS_IMETHODIMP
mozCellMLVariable::WriteRuntimeIVAssignment(nsACString& code)
{
  nsAutoString initialStr;

  if (seenRuntimeIVAssignment)
    return NS_OK;

  seenRuntimeIVAssignment = PR_TRUE;

  nsresult ret = variable->GetAttribute(NS_LITERAL_STRING("initial_value"),
                                        initialStr);
  if (ret != NS_OK ||
      initialStr.IsVoid())
    return NS_ERROR_NOT_AVAILABLE;

  nsCString initialCStr;
  CopyUTF16toUTF8(initialStr, initialCStr);

  nsXPIDLString componentName;
  GetComponentName(getter_Copies(componentName));
  nsAutoString inp;
  mozcellml->GetImportNumberPrefix(variable, nsnull, inp);

  // Use the model to resolve the initial value...
  nsCOMPtr<mozCellMLVariable> v
    = mozcellml->FindVariable(inp, componentName, initialStr);
  if (v == nsnull)
    return NS_OK;

  nsCAutoString cvn;
  GetVariableName(cvn);
  code += NS_LITERAL_CSTRING("  ") + cvn + NS_LITERAL_CSTRING(" = ");
  v->GetVariableName(cvn);
  code += cvn + NS_LITERAL_CSTRING(";\n");
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLVariable::GetInitialValue(double* initial)
{
  char* initialEnd;
  nsAutoString initialStr;
  nsresult ret = variable->GetAttribute(NS_LITERAL_STRING("initial_value"), initialStr);
  if (ret != NS_OK ||
      initialStr.IsVoid())
    return NS_ERROR_NOT_AVAILABLE;
  nsCString initialCStr;
  CopyUTF16toUTF8(initialStr, initialCStr);
  *initial = PR_strtod(initialCStr.get(), &initialEnd);
  if (initialEnd == initialCStr.get())
  {
    *initial = 1234.56;
    nsXPIDLString componentName;
    GetComponentName(getter_Copies(componentName));
    /* Don't have infinite recursion.
     * XXX threadsafety(but not just here)
     */
    if (AlreadyInIVCall)
    {
      nsXPIDLString variableName;
      GetName(getter_Copies(variableName));
      nsCAutoString cname, ccomponent;
      CopyUTF16toUTF8(variableName, cname);
      CopyUTF16toUTF8(componentName, ccomponent);
      COMPLAIN(mozcellml, (T(mozcellml, "infiniteIVCycle"), cname.get(), ccomponent.get()))
    }
    AlreadyInIVCall = PR_TRUE;
    nsAutoString inp;
    mozcellml->GetImportNumberPrefix(variable, nsnull, inp);
    // Use the model to resolve the initial value...
    nsCOMPtr<mozCellMLVariable> v
      = mozcellml->FindVariable(inp, componentName, initialStr);
    if (v == nsnull)
    {
      nsCAutoString ccomponent;
      CopyUTF16toUTF8(componentName, ccomponent);
      AlreadyInIVCall = PR_FALSE;
      COMPLAIN(mozcellml, (T(mozcellml, "invalidIVName"), initialCStr.get(),
                           ccomponent.get()))
    }
    ret = v->GetInitialValue(initial);
    AlreadyInIVCall = PR_FALSE;
    return ret;
  }
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLVariable::GetUnits(PRUnichar** ounits)
{
  nsAutoString units;
  nsresult ret = variable->GetAttribute(NS_LITERAL_STRING("units"), units);
  *ounits = ToNewUnicode(units);
  return ret;
}

nsresult
mozCellMLVariable::GetVariableName(nsACString& vn)
{
  vn = varname;
  return NS_OK;
}

nsresult
mozCellMLVariable::SetVariableName(const nsACString& vn)
{
  varname = vn;
  return NS_OK;
}

nsresult
mozCellMLVariable::GetVariableRateName(nsACString& vn)
{
  vn = varRateName;
  return NS_OK;
}

nsresult
mozCellMLVariable::SetVariableRateName(const nsACString& vn)
{
  varRateName = vn;
  return NS_OK;
}

nsresult
mozCellMLVariable::SetCode(PRInt32 code)
{
  varcode = code;
  return NS_OK;
}

PRInt32
mozCellMLVariable::GetCode()
{
  return varcode;
}

NS_IMETHODIMP
mozCellMLVariable::SetInitialValue(double value)
{
  nsPrintfCString ivCString("%e", value);
  nsAutoString ivString;
  CopyUTF8toUTF16(ivCString, ivString);
  return variable->SetAttribute(NS_LITERAL_STRING("initial_value"), ivString);
}

NS_IMETHODIMP
mozCellMLVariable::GetPublicInterface(
                                      PRUnichar** iface
                                     )
{
  nsAutoString ifacel;
  nsresult ret = variable->GetAttribute(NS_LITERAL_STRING("public_interface"), ifacel);
  if (ret != NS_OK || ifacel.IsVoid())
      return NS_ERROR_FAILURE;
  *iface = ToNewUnicode(ifacel);
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLVariable::GetPrivateInterface(
                                       PRUnichar** iface
                                      )
{
  nsAutoString ifacel;
  nsresult ret = variable->GetAttribute(NS_LITERAL_STRING("private_interface"), ifacel);
  if (ret != NS_OK || ifacel.IsVoid())
    return NS_ERROR_FAILURE;
  *iface = ToNewUnicode(ifacel);
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLVariable::IsFlagged(PRUint32 flags, PRBool *isFlagged)
{
  *isFlagged = PR_FALSE;
  if (flags & mozICellML::SHOULD_HAVE_INITIAL_VALUE)
  {
    if (type == mozICellMLVariable::SIMPLE ||
        type == mozICellMLVariable::BOUND)
      return NS_OK;
  }

  if (
      flags &
      (mozICellML::HAS_INITIAL_VALUE|mozICellML::NO_INITIAL_VALUE)
     )
  {
    double initial;
    nsresult res = GetInitialValue(&initial);
    if (res==NS_OK && (flags & mozICellML::NO_INITIAL_VALUE))
      return NS_OK;
    else if (res != NS_OK && (flags & mozICellML::HAS_INITIAL_VALUE))
      return NS_OK;
  }

  if (
      flags &
      (mozICellML::HAS_PUBLIC_INTERFACE|mozICellML::NO_PUBLIC_INTERFACE|
       mozICellML::NO_INTERFACE_IN)
     )
  {
    nsXPIDLString pubinterface, privinterface;
    nsresult res = GetPublicInterface(getter_Copies(pubinterface));
    nsresult res2 = GetPrivateInterface(getter_Copies(privinterface));
    if (res==NS_OK && (flags & mozICellML::NO_PUBLIC_INTERFACE))
      return NS_OK;
    else if (res != NS_OK && (flags & mozICellML::HAS_PUBLIC_INTERFACE))
      return NS_OK;
    else if ((flags & mozICellML::NO_INTERFACE_IN) &&
             ((res == NS_OK &&
               pubinterface.Equals(NS_LITERAL_STRING("in"))) ||
              (res2 == NS_OK &&
               privinterface.Equals(NS_LITERAL_STRING("in")))))
      return NS_OK;
  }
  *isFlagged = PR_TRUE;
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLVariable::GetType(PRUint32* type)
{
  *type = this->type;
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLVariable::GetImportNumber(PRUint32* importNumber)
{
  *importNumber = impNo;
  return NS_OK;
}

void
mozCellMLVariable::BoundInformSiblings(
    mozCellML* mcm,
    nsInterfaceHashtable<nsStringHashKey,mozCellMLVariableResolution>&
      allRes,
    double multiplier,
    double offset
                                      )
{
  nsXPIDLString compName;
  nsAutoString inp;
  GetComponentName(getter_Copies(compName));
  mcm->GetImportNumberPrefix(variable, nsnull, inp);

  nsCOMPtr<nsIDOMNode> parentNode;
  variable->GetParentNode(getter_AddRefs(parentNode));
  nsCOMPtr<nsIDOMElement> parentEl(do_QueryInterface(parentNode));
  nsCOMPtr<nsIDOMNodeList> nl;
  parentEl->GetChildNodes(getter_AddRefs(nl));
  nsCOMPtr<nsIArray> allSiblings;
  FilterNodeListByNameAndTypeNS2(nl, getter_AddRefs(allSiblings),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("variable")
                                 );
  PRUint32 i, l;
  allSiblings->GetLength(&l);
  nsCOMPtr<nsISimpleEnumerator> e;
  allSiblings->Enumerate(getter_AddRefs(e));
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<nsISupports> varSupports;
    e->GetNext(getter_AddRefs(varSupports));
    nsCOMPtr<nsIDOMElement> varEl(do_QueryInterface(varSupports));
    nsAutoString varName;
    varEl->GetAttribute(NS_LITERAL_STRING("name"), varName);
    nsCOMPtr<mozCellMLVariableResolution> res;
    allRes.Get(inp + compName + NS_LITERAL_STRING(":") + varName,
               getter_AddRefs(res));
    if (res == nsnull)
    {
#ifdef DEBUG_MODE
      nsAutoString resName = inp + compName + NS_LITERAL_STRING(":") + varName;
      nsCAutoString cresName;
      CopyUTF16toUTF8(resName, cresName);
      printf("Can't find resolution %s\n", cresName.get());
#endif
      continue;
    }
    res->SetBound(multiplier, offset);
  }
}

NS_IMPL_ISUPPORTS0(mozCellMLVariablePromise)

mozCellMLVariablePromise::mozCellMLVariablePromise(
                                                   nsAString& party1,
                                                   nsAString& party2
                                                  )
{
  nsCAutoString cparty1, cparty2;
  CopyUTF16toUTF8(party1, cparty1);
  CopyUTF16toUTF8(party2, cparty2);
#ifdef DEBUG_MODE
  printf("Creating promise for %s <=> %s\n",
         cparty1.get(), cparty2.get());
#endif
  parties.AppendString(party1);
  partiesFrom.AppendString(party2);
}

void
mozCellMLVariablePromise::AddParty(nsAString& party,
                                   nsAString& partyFrom)
{
  if (redundantBy != nsnull)
    redundantBy->AddParty(party, partyFrom);
  parties.AppendString(party);
  partiesFrom.AppendString(partyFrom);
}

nsStringArray&
mozCellMLVariablePromise::SetRedundant(mozCellMLVariablePromise* r,
                                       nsStringArray** ret2)
{
  nsStringArray* ret;
  if (redundantBy != nsnull)
    ret = &redundantBy->SetRedundant(r, ret2);
  else
  {
    ret = &parties;
    *ret2 = &partiesFrom;
  }
  redundantBy = r;
  nsStringArray& retref = *ret;
  return retref;
}

void
mozCellMLVariablePromise::MergeIn(mozCellMLVariablePromise* merge)
{
  if (redundantBy != nsnull)
    redundantBy->MergeIn(merge);
  else
  {
    nsStringArray* mergeArray2;
    nsStringArray& mergeArray = merge->SetRedundant(this, &mergeArray2);
    PRUint32 i, l;
    nsAutoString party, partyFrom;
    for (i = 0, l = mergeArray.Count(); i < l; i++)
    {
      mergeArray.StringAt(i, party);
      mergeArray2->StringAt(i, partyFrom);
      parties.AppendString(party);
      partiesFrom.AppendString(partyFrom);
    }
  }
}

nsresult
mozCellMLVariablePromise::PartySolved(
                                      mozCellML* model,
                                      nsAString& party,
                                      mozCellMLVariableResolution* var,
                                      nsInterfaceHashtable<nsStringHashKey,
                                                           mozCellMLVariable>&
                                        nonMainVariables,
                                      nsInterfaceHashtable<nsStringHashKey,
                                                           mozCellMLVariableResolution>& restable,
                                      nsCOMArray<mozCellMLVariableResolution>& allRes
                                     )
{
  PRInt32 i;
  for (i = 0; i < parties.Count(); i++)
  {
    nsAutoString party2;
    parties.StringAt(i, party2);

    if (party.Equals(party2))
    {
      partiesFrom.StringAt(i, party2);
    }
    else
    {
      partiesFrom.StringAt(i, party2);
      if (!party.Equals(party2))
        continue;
      parties.StringAt(i, party2);
    }

    nsCOMPtr<mozCellMLVariableResolution> var2;
    nsCOMPtr<mozCellMLVariable> nmvar;
    nonMainVariables.Get(party2, getter_AddRefs(nmvar));

    nsresult rv;
    rv = var->GetConversion(model, nmvar, allRes, getter_AddRefs(var2));
    if (var2 == nsnull)
      printf("Got null conversion!!!\n");
    NS_ENSURE_SUCCESS(rv, rv);
    restable.Put(party2, var2);
    parties.RemoveStringAt(i);
    partiesFrom.RemoveStringAt(i);
    rv = PartySolved(model, party2, var2, nonMainVariables, restable,
                     allRes);
    NS_ENSURE_SUCCESS(rv, rv);
    i--;
  }
  // Look for all parties which directly include this one.
  return NS_OK;
}


nsresult
mozCellMLVariablePromise::Resolve(
                                  mozCellML* model,
                                  nsInterfaceHashtable<nsStringHashKey,
                                                       mozCellMLVariable>&
                                    nonMainVariables,
                                  nsInterfaceHashtable<nsStringHashKey,
                                                       mozCellMLVariableResolution>& restable,
                                  nsInterfaceHashtable<nsStringHashKey,
                                                       mozCellMLVariable>& vartable,
                                  nsCOMArray<mozCellMLVariableResolution>& allRes
                                 )
{
  if (redundantBy != nsnull)
    return NS_OK;
  nsAutoString party;
  nsCOMPtr<mozCellMLVariableResolution> var;
  PRUint32 i, j, n;
  PRBool got_solution = PR_FALSE;
  for (i = 0; i < (PRUint32)parties.Count(); i++)
  {
    parties.StringAt(i, party);

    restable.Get(party, getter_AddRefs(var));
    if (var == nsnull)
      continue;
    // We just found a solution. Recursively solve...

    if (!got_solution)
      for (j = 0, n = parties.Count(); j < n; j++)
      {
        nsAutoString party2;
        parties.StringAt(j, party2);
        vartable.Put(party2, var->GetVariable());
        partiesFrom.StringAt(j, party2);
        vartable.Put(party2, var->GetVariable());
      }
    got_solution = PR_TRUE;

    nsresult rv = PartySolved(model, party, var, nonMainVariables, restable,
                              allRes);
    i = 0;
    NS_ENSURE_SUCCESS(rv, rv);
  }
  for (i = 0; i < (PRUint32)parties.Count(); i++)
  {
    partiesFrom.StringAt(i, party);

    restable.Get(party, getter_AddRefs(var));
    if (var == nsnull)
      continue;
    // We just found a solution. Recursively solve...

    if (!got_solution)
      for (j = 0, n = parties.Count(); j < n; j++)
      {
        nsAutoString party2;
        parties.StringAt(j, party2);
        vartable.Put(party2, var->GetVariable());
        partiesFrom.StringAt(j, party2);
        vartable.Put(party2, var->GetVariable());
      }
    got_solution = PR_TRUE;

    nsresult rv = PartySolved(model, party, var, nonMainVariables, restable, allRes);
    i = 0;
    NS_ENSURE_SUCCESS(rv, rv);
  }
  if (parties.Count() == 0)
    return NS_OK;
  nsAutoString party1, party2;
  parties.StringAt(0, party1);
  partiesFrom.StringAt(0, party2);

  nsCAutoString cparty1, cparty2;
  CopyUTF16toUTF8(party1, cparty1);
  CopyUTF16toUTF8(party2, cparty2);
  COMPLAIN(model, (T(model, "unsolvedConnection"), cparty1.get(),
                   cparty2.get()));
}

NS_IMPL_ISUPPORTS0(mozCellMLVariableResolution)

mozCellMLVariableResolution
  ::mozCellMLVariableResolution(mozCellMLVariable* var,
                                mozCellMLVariable* nmvar,
                                double a, double b)
    : variable(var), nmvariable(nmvar), mup(a), constv(b)
{
}

mozCellMLVariableResolution::~mozCellMLVariableResolution()
{
}

void
mozCellMLVariableResolution::BoundInformSiblings(
                                                 mozCellML* mcm,
                                                 nsInterfaceHashtable<nsStringHashKey,
                                                 mozCellMLVariableResolution>& allRes)
{
  PRUint32 type;
  variable->GetType(&type);
  if (type == mozICellMLVariable::BOUND)
    nmvariable->BoundInformSiblings(mcm, allRes, mup, constv);
}

void
mozCellMLVariableResolution::GetVariableRHS(nsACString& rhsOut)
{
  if (mup == 1.0 && constv == 0.0)
  {
    variable->GetVariableName(rhsOut);
    return;
  }
  if (mup == 1.0)
  {
    nsCAutoString tmp;
    variable->GetVariableName(tmp);
    rhsOut.Assign("(");
    rhsOut.Append(tmp);
    rhsOut.Append(" + ");
    nsPrintfCString ps("%e)", constv);
    rhsOut.Append(ps);
    return;
  }
  if (constv == 0.0)
  {
    nsCAutoString tmp;
    variable->GetVariableName(tmp);
    rhsOut.Assign("(");
    rhsOut.Append(tmp);
    rhsOut.Append(" * ");
    nsPrintfCString ps("%e", mup);
    rhsOut.Append(ps);
    rhsOut.Append(")");
    return;
  }
  nsCAutoString tmp;
  variable->GetVariableName(tmp);
  rhsOut.Assign("(");
  rhsOut.Append(tmp);
  rhsOut.Append("*");
  nsPrintfCString ps("%e", mup);
  rhsOut.Append(ps);
  rhsOut.Append(" + ");
  nsPrintfCString ps2("%e", constv);
  rhsOut.Append(ps2);
  rhsOut.Append(")");
}

void
mozCellMLVariableResolution::GetAssignCode(
                                           PRUint32 i,
                                           nsACString& rhsIn,
                                           nsACString& codeOut)
{
  nsPrintfCString cps("  x[%u] = (", i);
  codeOut.Append(cps);
  if (mup == 1.0)
  {
    codeOut.Append(rhsIn);
    if (constv != 0.0)
    {
      codeOut.Append(" + ");
      nsPrintfCString ps("%e", constv);
      codeOut.Append(ps);
    }
  }
  else if (constv == 0.0)
  {
    codeOut.Append(rhsIn);
    codeOut.Append(" * ");
    // XXX - what if mup == 0?
    nsPrintfCString ps("%e", (1.0/mup));
    codeOut.Append(ps);
  }
  else
  {
    codeOut.Append("(");
    codeOut.Append(rhsIn);
    codeOut.Append(" + ");
    nsPrintfCString ps("%e", constv);
    codeOut.Append(ps);
    codeOut.Append(") * ");
    // XXX - what if mup == 0?
    nsPrintfCString ps2("%e", (1.0/mup));
    codeOut.Append(ps2);
  }
  codeOut.Append(");");
}

void
mozCellMLVariableResolution::GetDiffAssignCode(
                                               PRUint32 i,
                                               nsACString& rhsIn,
                                               nsACString& codeOut,
                                               PRUint32 degree)
{
  nsPrintfCString cps("  dx[%u] = ", i);
  codeOut.Append(cps);
  double mb = 1 / (mup * pow(bmup, degree));
  if (mb == 1.0)
  {
    codeOut.Append(rhsIn);
  }
  else
  {
    codeOut.Append("(");
    codeOut.Append(rhsIn);
    codeOut.Append(") * ");
    // XXX - what if mup == 0?
    nsPrintfCString ps2("%e", mb);
    codeOut.Append(ps2);
  }

  codeOut.Append(";");
}

#define IFOP(n) else if (lname.Equals(NS_LITERAL_STRING(#n)))
#define FUNCCONST(n,v) IFOP(n)\
                         code += NS_LITERAL_CSTRING(v);

#define ROOTCONST(n,v)  else if (name.Equals(NS_LITERAL_STRING(#n)))\
                          code += NS_LITERAL_CSTRING(v);

#define FUNCOP1(n,f) IFOP(n){\
                      if (numSubs != 2)\
                      {\
                        COMPLAIN(model, (T(model, "applyWrongChildren2"), #n, numSubs));\
                        return NS_ERROR_FAILURE;\
                      }\
                      code += NS_LITERAL_CSTRING(#f "("); \
                      e->GetNext(getter_AddRefs(opSupports)); \
                      op = do_QueryInterface(opSupports); \
                      ret = GetComputationString(model, op, allVariables, importNumbers, allUnits, code, functionStack); \
                      if (ret != NS_OK) \
                        return ret; \
                      code += NS_LITERAL_CSTRING(")"); \
                    }
#define FUNCOP2(n,f) IFOP(n){\
                      if (numSubs != 3)\
                      {\
                        COMPLAIN(model, (T(model, "applyWrongChildren3"),  #n, numSubs));\
                        return NS_ERROR_FAILURE;\
                      }\
                      code += NS_LITERAL_CSTRING(#f "(");\
                      e->GetNext(getter_AddRefs(opSupports)); \
                      op = do_QueryInterface(opSupports); \
                      ret = GetComputationString(model, op, allVariables, importNumbers, allUnits, code, functionStack); \
                      if (ret != NS_OK) \
                        return ret; \
                      code += ", "; \
                      e->GetNext(getter_AddRefs(opSupports)); \
                      op = do_QueryInterface(opSupports); \
                      ret = GetComputationString(model, op, allVariables, importNumbers, allUnits, code, functionStack); \
                      if (ret != NS_OK) \
                        return ret; \
                      code += NS_LITERAL_CSTRING(")"); \
                     }
#define FUNCROP1(n,f) IFOP(n){\
                      if (numSubs != 2)\
                      {\
                        COMPLAIN(model, (T(model, "applyWrongChildren2"), #n, numSubs));\
                        return NS_ERROR_FAILURE;\
                      }\
                      code += NS_LITERAL_CSTRING("1 / " #f "(");\
                      e->GetNext(getter_AddRefs(opSupports)); \
                      op = do_QueryInterface(opSupports); \
                      ret = GetComputationString(model, op, allVariables, importNumbers, allUnits, code, functionStack); \
                      if (ret != NS_OK) \
                        return ret; \
                      code += NS_LITERAL_CSTRING(")"); \
                    }
#define FUNCOPR1(n,f) IFOP(n){\
                      if (numSubs != 2)\
                      {\
                        COMPLAIN(model, (T(model, "applyWrongChildren2"), #n, numSubs));\
                        return NS_ERROR_FAILURE;\
                      }\
                      code += NS_LITERAL_CSTRING(#f "(1 / (");\
                      e->GetNext(getter_AddRefs(opSupports)); \
                      op = do_QueryInterface(opSupports); \
                      ret = GetComputationString(model, op, allVariables, importNumbers, allUnits, code, functionStack); \
                      if (ret != NS_OK) \
                        return ret; \
                      code += NS_LITERAL_CSTRING("))"); \
                    }
#define COPERN(n,o) IFOP(n){\
                    if (numSubs < 2)\
                    {\
                      COMPLAIN(model, (T(model, "applyWrongChildren2P"), #n, numSubs));\
                      return NS_ERROR_FAILURE;\
                    }\
                    code += NS_LITERAL_CSTRING("(");\
                    e->GetNext(getter_AddRefs(opSupports)); \
                    op = do_QueryInterface(opSupports); \
                    ret = GetComputationString(model, op, allVariables, importNumbers, allUnits, code, functionStack); \
                    if (ret != NS_OK) \
                      return ret; \
                    code += NS_LITERAL_CSTRING(")"); \
                    for (PRUint32 x = 2; x < numSubs; x++) \
                    { \
                      code += NS_LITERAL_CSTRING(o "("); \
                      e->GetNext(getter_AddRefs(opSupports)); \
                      op = do_QueryInterface(opSupports); \
                      ret = GetComputationString(model, op, allVariables, importNumbers, allUnits, code, functionStack); \
                      if (ret != NS_OK) \
                        return ret; \
                      code += NS_LITERAL_CSTRING(")"); \
                    } \
                  }
#define COPER12(n,o) IFOP(n){\
                    if (numSubs != 2 && numSubs != 3)\
                    {\
                      COMPLAIN(model, (T(model, "applyWrongChildren23"), #n, numSubs));\
                      return NS_ERROR_FAILURE;\
                    }\
                    if (numSubs == 2)\
                      code += NS_LITERAL_CSTRING(o);\
                    code += NS_LITERAL_CSTRING("(");\
                    e->GetNext(getter_AddRefs(opSupports)); \
                    op = do_QueryInterface(opSupports); \
                    ret = GetComputationString(model, op, allVariables, importNumbers, allUnits, code, functionStack); \
                    if (ret != NS_OK) \
                      return ret; \
                    code += NS_LITERAL_CSTRING(")"); \
                    if (numSubs == 3)\
                    { \
                      code += NS_LITERAL_CSTRING(o "("); \
                      e->GetNext(getter_AddRefs(opSupports)); \
                      op = do_QueryInterface(opSupports); \
                      ret = GetComputationString(model, op, allVariables, importNumbers, allUnits, code, functionStack); \
                      if (ret != NS_OK) \
                        return ret; \
                      code += NS_LITERAL_CSTRING(")"); \
                    } \
                  }
#define COPER1(n,o) IFOP(n){\
                    if (numSubs != 2)\
                    {\
                      COMPLAIN(model, (T(model, "applyWrongChildren2"), #n, numSubs));\
                      return NS_ERROR_FAILURE;\
                    }\
                    code += NS_LITERAL_CSTRING("!(");\
                    e->GetNext(getter_AddRefs(opSupports)); \
                    op = do_QueryInterface(opSupports); \
                    ret = GetComputationString(model, op, allVariables, importNumbers, allUnits, code, functionStack); \
                    if (ret != NS_OK) \
                      return ret; \
                    code += NS_LITERAL_CSTRING(")"); \
                  }
#define FUNCDEFAULT(n,f,d,m,o) IFOP(n){\
      if (numSubs < 2 || numSubs > 3)\
      {\
        COMPLAIN(model, (T(model, "applyWrongChildren23"), #n, numSubs));\
        return NS_ERROR_FAILURE;\
      }\
      nsCOMPtr<nsIDOMElement> inside;\
      if (numSubs == 2)\
      {\
        nsCOMPtr<nsISupports> insideSupports; \
        e->GetNext(getter_AddRefs(insideSupports));\
        inside = do_QueryInterface(insideSupports); \
        code += #f "(";\
        ret = GetComputationString(model, inside, allVariables, importNumbers, allUnits, code, functionStack);\
        if (ret != NS_OK) \
          return ret; \
        code += ", " #d ")"; \
      }\
      else\
      {\
        nsCOMPtr<nsISupports> p1Supports, p2Supports; \
        e->GetNext(getter_AddRefs(p1Supports)); \
        e->GetNext(getter_AddRefs(p2Supports)); \
        nsCOMPtr<nsIDOMElement> p1(do_QueryInterface(p1Supports)), \
                                p2(do_QueryInterface(p2Supports)), degV; \
        nsAutoString nm1, nm2;\
        p1->GetLocalName(nm1);\
        p2->GetLocalName(nm2);\
        if (nm1.Equals(NS_LITERAL_STRING(#o)))\
        {\
          inside = p2;\
          degV = p1;\
        }\
        else if (nm2.Equals(NS_LITERAL_STRING(#o)))\
        {\
          inside = p1;\
          degV = p2;\
        }\
        else\
        {\
          COMPLAIN(model, (T(model, "malformedFuncInvalidChild"), #n, #o));\
          return NS_ERROR_FAILURE;\
        }\
        nsCOMPtr<nsIDOMNodeList> degSubnodesAll;\
        nsCOMPtr<nsIArray> degSubnodes;\
        degV->GetChildNodes(getter_AddRefs(degSubnodesAll));\
        FilterNodeListByTypeNS(degSubnodesAll, getter_AddRefs(subnodes),\
                               nsIDOMNode::ELEMENT_NODE, MATHML_NS);\
        PRUint32 numDegs;\
        degSubnodes->GetLength(&numDegs);\
        if (numDegs != 1)\
        {\
          COMPLAIN(model, (T(model, "malformedFuncWrongOverride"), #n, #m, numDegs));\
          return NS_ERROR_FAILURE;\
        }\
        nsCOMPtr<nsISimpleEnumerator> de;\
        degSubnodes->Enumerate(getter_AddRefs(de));\
        nsCOMPtr<nsISupports> ddeSupports; \
        de->GetNext(getter_AddRefs(ddeSupports)); \
        nsCOMPtr<nsIDOMElement> dde(do_QueryInterface(ddeSupports)); \
        code += #f "(";\
        ret = GetComputationString(model, inside, allVariables, importNumbers, allUnits, code, functionStack);\
        if (ret != NS_OK) \
          return ret; \
        code += ", " m "(";\
        ret = GetComputationString(model, dde, allVariables, importNumbers, allUnits, code, functionStack);\
        if (ret != NS_OK) \
          return ret; \
        code += "))";\
      } \
                             }

void
mozCellMLVariable::GetUnits(mozCellMLUnits** units)
{
  *units = this->units;
  NS_ADDREF(*units);
}

nsresult
mozCellMLVariableResolution::
  GetConversion(
                mozCellML* model,
                mozCellMLVariable* target,
                nsCOMArray<mozCellMLVariableResolution>& allResolutions,
                mozCellMLVariableResolution** result)
{
  // See if the target units is the same as the variable...
  nsCOMPtr<mozCellMLUnits> unitsDefTarg, unitsDefVar;
  target->GetUnits(getter_AddRefs(unitsDefTarg));
  variable->GetUnits(getter_AddRefs(unitsDefVar));
#if 0
  if (unitsDefTarg == unitsDefVar ||
      unitsDefTarg->Equals(unitsDefVar))
  {
    NS_ADDREF(this);
    allResolutions.AppendObject(*result);
    *result = this;
    return NS_OK;
  }
#endif
  // Not exactly the same. See if at least the dimensions are consistant...
  if (!unitsDefTarg->Equivalent(unitsDefVar))
  {
    nsCAutoString cfromComponent, ctoComponent;
    nsCAutoString cfromVariable, ctoVariable;

    nsXPIDLString fromVariable, fromComponent;
    variable->GetName(getter_Copies(fromVariable));
    variable->GetComponentName(getter_Copies(fromComponent));
    CopyUTF16toUTF8(fromVariable, cfromVariable);
    CopyUTF16toUTF8(fromComponent, cfromComponent);

    nsXPIDLString toVariable, toComponent;
    target->GetName(getter_Copies(toVariable));
    target->GetComponentName(getter_Copies(toComponent));
    CopyUTF16toUTF8(toVariable, ctoVariable);
    CopyUTF16toUTF8(toComponent, ctoComponent);

    COMPLAIN(model, (T(model, "mapInconsistantDimensions"),
                     cfromVariable.get(),
                     cfromComponent.get(),
                     ctoVariable.get(),
                     ctoComponent.get()));
  }

  // To resolve a variable, we need to look at both what it is being converted
  // from, and what it is being converted to.
  double nmup = unitsDefTarg->GetMultiplier() / unitsDefVar->GetMultiplier();
  double noff = unitsDefTarg->GetOffset() - unitsDefVar->GetOffset();

  *result = new mozCellMLVariableResolution(variable, target, nmup, noff);
  NS_ADDREF(*result);
  allResolutions.AppendObject(*result);
  return NS_OK;
}

void
mozCellMLVariableResolution::SetBound(double mup, double off)
{
  bmup = mup;
  bconstv = off;
}
