/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.0/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the mozCellML code.
 *
 * The Initial Developer of the Original Code is
 * The Bioengineering Institute, University of Auckland.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrew Miller <ak.miller@auckland.ac.nz>
 *
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 */
#include <nsIGenericFactory.h>
#include <nsISupports.h>
#include <nsCOMPtr.h>
#include <nsIProperties.h>
#include <nsDirectoryService.h>
#include <nsIComponentRegistrar.h>
#include <nspr/prenv.h>
#include <nspr/plstr.h>

class mozCellMLPreparer
  : public nsISupports
{
public:
  mozCellMLPreparer();
  NS_DECL_ISUPPORTS
protected:
  virtual ~mozCellMLPreparer() {}
};

NS_IMPL_ISUPPORTS0(mozCellMLPreparer)

static void
DoPreparations(nsIFile* pdir)
{
#ifdef XP_WIN
  char* envstr;
  nsAutoString path;
  nsCAutoString cpath;

  pdir->GetPath(path);
  CopyUTF16toUTF8(path, cpath);
  envstr = PR_GetEnv("PATH");
  // On Win32, libraries are picked up from the same directory as the
  // executable, and so Mozilla doesn't put the XPCOM/NSPR executables into the
  // path. We need this for the integrator/fitter to work...
  
  nsCOMPtr<nsIProperties>
    dsp(do_GetService(NS_DIRECTORY_SERVICE_CONTRACTID));
  if (dsp == (nsIProperties*)nsnull)
  {
    return;
  }
  nsCOMPtr<nsIFile> cdir, cdir2;
  nsresult rv =
    dsp->Get("ComsD", NS_GET_IID(nsIFile), getter_AddRefs(cdir));
  if (cdir == nsnull)
  {
    printf("Can't get hold of the components directory.\n");
    return;
  }
  cdir->GetParent(getter_AddRefs(cdir2));
  if (cdir2 == nsnull)
  {
    printf("Can't get hold of the components directory parent.\n");
    return;
  }

  cdir2->GetPath(path);
  nsCAutoString cdpath;
  CopyUTF16toUTF8(path, cdpath);
  cpath += NS_LITERAL_CSTRING(";");
  cpath += cdpath;
  fflush(stdout);

  nsCAutoString cpt(NS_LITERAL_CSTRING("PATH="));
  if (envstr == nsnull)
  {
    cpt += cpath;
    PR_SetEnv(PL_strdup(cpt.get()));
  }
  else
  {
    cpt += cpath;
    cpt += NS_LITERAL_CSTRING(";");
    cpt += envstr;
    PR_SetEnv(PL_strdup(cpt.get()));
  }
#else
  // We need to copy the files to the Mozilla directory...
  nsCOMPtr<nsIProperties>
    dsp(do_GetService(NS_DIRECTORY_SERVICE_CONTRACTID));
  if (dsp == nsnull)
  {
    printf("Directory service could not be retrieved.\n");
    return;
  }
  nsCOMPtr<nsIFile> cdir, cdir1;
  nsresult rv =
    dsp->Get("ComsD", NS_GET_IID(nsIFile), getter_AddRefs(cdir));
  if (cdir == nsnull)
  {
    printf("Component directory could not be retrieved.\n");
    return;
  }
  cdir->GetParent(getter_AddRefs(cdir1));
  if (cdir1 == nsnull)
  {
    printf("Component directory parent could not be retrieved.\n");
    return;
  }

  // We want to copy the files into cdir...
  nsCOMPtr<nsIFile> fname;
  cdir1->Clone(getter_AddRefs(fname));
  fname->Append(NS_LITERAL_STRING("libomniORB4.so.0"));
  PRBool e;
  fname->Exists(&e);
  if (e)
    return;

  pdir->Clone(getter_AddRefs(fname));
  fname->Append(NS_LITERAL_STRING("libomniORB4.so.0"));
  fname->CopyTo(cdir1, NS_LITERAL_STRING(""));

  pdir->Clone(getter_AddRefs(fname));
  fname->Append(NS_LITERAL_STRING("libomnithread.so.3"));
  fname->CopyTo(cdir1, NS_LITERAL_STRING(""));

  pdir->Clone(getter_AddRefs(fname));
  fname->Append(NS_LITERAL_STRING("libxpcorba.so"));
  fname->CopyTo(cdir1, NS_LITERAL_STRING(""));
#endif
}

mozCellMLPreparer::mozCellMLPreparer()
{
#ifdef XP_WIN
  setvbuf(stdout, nsnull, _IONBF, 0);
  setvbuf(stderr, nsnull, _IONBF, 0);
#endif

  nsCOMPtr<nsIProperties>
    dsp(do_GetService(NS_DIRECTORY_SERVICE_CONTRACTID));
  if (dsp == nsnull)
  {
    return;
  }
  nsCOMPtr<nsIFile> pdir;
  nsresult rv =
    dsp->Get("ProfD", NS_GET_IID(nsIFile), getter_AddRefs(pdir));
  /* Don't crash if it doesn't work... */
  if (pdir == nsnull)
  {
    return;
  }
  pdir->Append(NS_LITERAL_STRING("extensions"));
  pdir->Append(NS_LITERAL_STRING("{a0d0a9af-bb3d-46c1-a470-8800b0b3f3e2}"));

  DoPreparations(pdir);
}

static NS_METHOD
mozCellMLPreparerRegistration(
                      nsIComponentManager *aCompMgr,
                      nsIFile *aPath,
                      const char *registryLocation,
                      const char *componentType,
                      const nsModuleComponentInfo *info
                     )
{
  nsCOMPtr<nsIFile> tdir, pdir;
  aPath->Clone(getter_AddRefs(pdir));
  pdir->GetParent(getter_AddRefs(tdir));
  tdir->GetParent(getter_AddRefs(pdir));
  DoPreparations(pdir);

  
  // It is now safe to load the remaining components...
  nsCOMPtr<nsIComponentRegistrar> ncr;
  aCompMgr->QueryInterface(NS_GET_IID(nsIComponentRegistrar),
                           getter_AddRefs(ncr));
  pdir->Append(NS_LITERAL_STRING("components_needcorba"));
  ncr->AutoRegister(pdir);
  return NS_OK;
}

NS_GENERIC_FACTORY_CONSTRUCTOR(mozCellMLPreparer)

#define MOZCELLML_PREPARER_CID                       \
{ /* 9b325a02-238b-4fd6-8655-1a568e0ea9ac */         \
    0x9b325a02,                                      \
    0x238b,                                          \
    0x4fd6,                                          \
    {0x86, 0x55, 0x1a, 0x56, 0x8e, 0x0e, 0xa9, 0xac} \
}

static const nsModuleComponentInfo components[] = 
{
  {
    "mozCellMLPreparer",
    MOZCELLML_PREPARER_CID,
    "@physiome.org/mozCellMLPreparer;1",
    mozCellMLPreparerConstructor,
    mozCellMLPreparerRegistration
  }
};

NS_IMPL_NSGETMODULE(MCPModule,components)
