#!/usr/bin/python
# Copyright (C) 2001,  Christof Meerwald <cmeerw@web.de>
# http://xpcom2corba.sunsite.dk/

# XPCOM-CORBA is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.

import os, string, sys

from omniidl import idlast, idlvisitor, idlutil, idltype, output
from xpcom_tmpl import *
import xpcorba_type
convert_scoped2flat = xpcorba_type.convert_scoped2flat


REPOID_ISUPPORTS = 'DCE:00000000-0000-0000-c000-000000000046:1'


class XPCOM_Proxy_Visitor(idlvisitor.AstVisitor):
    def __init__(self):
        self._header = None
        self._impl = None
        self._toplevel = []
        self._included = ['xpcom.idl']


    def visitAST(self, node):
        directory, basename = os.path.split(node.file())
        if string.lower(basename[-4:]) == '.idl':
            basename = basename[:-4]

        self._impl = output.Stream(open(basename + '_xpcom.cc', 'w'), 2)
        self._header = output.Stream(open(basename + '_xpcom.h', 'w'), 2)

        self._header.out(TMPL_proxy_h_file_header, basename=basename,
                         guardname='HEADER__' + string.upper(basename) + '_XPCOM__INCLUDED',
                         CORBAprefix='', CORBAsuffix='.hh',
                         XPCOMprefix='I', XPCOMsuffix='.h')

        self._impl.out(TMPL_proxy_c_file_header, basename=basename)

        for n in node.declarations():
            if n.mainFile():
                n.accept(self)
            else:
                filename = n.file()
                pos = string.rfind(filename, '/')
                if pos != -1:
                    filename = filename[pos + 1:]

                if not filename in self._included:
                    self._included.append(filename)

                    if filename[-4:] == '.idl':
                        filename = filename[:-4]

                    self._header.out('#include "@filename@_xpcom.h"\n#include "@filename@_corba.h"\n',
                                   filename=filename)

        for s, args in self._toplevel:
            apply(self._header.out, [s], args)
        self._toplevel = []

        self._header.out(TMPL_proxy_h_file_footer)

        self._impl = None
        self._header = None


    def visitModule(self, node):
        for n in node.definitions():
            if n.mainFile():
                n.accept(self)


    def visitTypedef(self, node):
        for decl in node.declarators():
            if isinstance(node.aliasType(), idltype.Sequence):
                childtype = node.aliasType().seqType().unalias()

                corba_name = idlutil.ccolonName(decl.scopedName())
                xpchild_type = xpcorba_type.get_type(childtype)
                xpcom_childname = xpchild_type.type_xpcom()

                if isinstance(childtype, idltype.String) or isinstance(childtype, idltype.WString):
                    self._toplevel.append((TMPL_proxy_h_sequence_string,
                                           {'corba_name': corba_name,
                                            'xpcom_childname': xpcom_childname}))
                elif isinstance(childtype, idltype.Declared) and \
                         childtype.kind() == idltype.tk_objref:
                    xpcom_childname = xpchild_type._intfname
                    corba_childname = xpchild_type._corbaname
                    self._toplevel.append((TMPL_proxy_h_sequence_objref,
                                           {'corba_name': corba_name,
                                            'xpcom_childname': xpcom_childname,
                                            'corba_childname': corba_childname}))
                else:
                    self._toplevel.append((TMPL_proxy_h_sequence,
                                           {'corba_name': corba_name,
                                            'xpcom_childname': xpcom_childname}))


    def visitStruct(self, node):
        corba_name = idlutil.ccolonName(node.scopedName())
        xpcom_name = convert_scoped2flat(node.scopedName())


        toplevel = [TMPL_proxy_h_struct_x2c_header]

        for member in node.members():
            identifier = member.declarators()[0].identifier()

            xpcom_type = xpcorba_type.get_type(member.memberType())
            toplevel.append('  {')
            toplevel.append('    ' + xpcom_type.xpcom_preinvoke('corba_obj.' + identifier, 0, 1))
            toplevel.append('    xpcom_obj->Get' + string.upper(identifier[0]) + identifier[1:] + '(' + xpcom_type.call_xpcom_param('corba_obj.' + identifier, 0, 1) + ');')
            toplevel.append('    ' + xpcom_type.xpcom_postinvoke('corba_obj.' + identifier, 1, 1))
            toplevel.append('  }')

        toplevel.append(TMPL_proxy_h_struct_x2c_footer)

        self._toplevel.append((string.join(toplevel, '\n'),
                               {'corba_name': corba_name,
                                'xpcom_name': xpcom_name}))



    def visitInterface(self, node):
        name = node.identifier()
        fqname = idlutil.ccolonName(node.scopedName())
        xpcomintfname = convert_scoped2flat(node.scopedName())

        if ((not node.inherits()) or
            ((len(node.inherits()) == 1) and
             (node.inherits()[0].repoId() == REPOID_ISUPPORTS))):
            if not node.inherits():
                supername = 'CORBA_IObject__XPCOM_Proxy'
                new_ref_count = 'NULL'
            else:
                supername = 'XPCOM_IObject__XPCOM_Proxy'
                new_ref_count = 'new PRInt32(1)'

            self._header.out(TMPL_proxy_h_intf_header_noinherit,
                             intfname=xpcomintfname, fqname=fqname,
                             supername=supername,
                             uppername=string.upper(xpcomintfname))

            self._impl.out(TMPL_proxy_c_intf,
                           intfname=xpcomintfname, fqname=fqname,
                           supername=supername, new_ref_count=new_ref_count)
        else:
            superintf = node.inherits()[0]
            supername = convert_scoped2flat(superintf.scopedName()) + '__XPCOM_Proxy'

            self._header.out(TMPL_proxy_h_intf_header_inherit,
                             intfname=xpcomintfname, fqname=fqname,
                             supername=supername,
                             uppername=string.upper(xpcomintfname))

            new_ref_count = 'NULL'
            inherits = node.inherits()
            while inherits:
                if inherits[0].repoId() != REPOID_ISUPPORTS:
                    self._header.out('  NS_FORWARD_@uppername@(@classname@)',
                                     uppername=string.upper(convert_scoped2flat(inherits[0].scopedName())),
                                     classname=convert_scoped2flat(superintf.scopedName()) + '__XPCOM_Proxy::')
                    inherits = inherits + inherits[0].inherits()
                else:
                    new_ref_count = 'new PRInt32(1)'
                del inherits[0]

            self._impl.out(TMPL_proxy_c_intf,
                           intfname=xpcomintfname, fqname=fqname,
                           supername=supername, new_ref_count=new_ref_count)

        self._header.inc_indent()

        for d in node.declarations():
            d.accept(self)

        for c in node.callables():
            c.accept(self)

        self._header.dec_indent()
        self._header.out(TMPL_proxy_h_intf_footer,
                         intfname=xpcomintfname, fqname=fqname)


    def visitAttribute(self, node):
        xpcom_type = xpcorba_type.get_type(node.attrType())

        paramspec = xpcom_type.decl_xpcom('_retval', 0, 1)

        corba_preinvoke = xpcom_type.corba_preinvoke('_retval', 0, 1)
        corba_postinvoke = xpcom_type.corba_postinvoke('_retval', 0, 1)

        for name in node.identifiers():
            capname = 'Get' + string.upper(name[0]) + name[1:]
            xpcomintfname = convert_scoped2flat(node.declarators()[0].scopedName()[:-1])

            self._header.out(TMPL_proxy_h_operation, name=name,
                             capname=capname, paramspec=paramspec)

            self._impl.out(TMPL_proxy_c_operation_header,
                           name=name, capname=capname,
                           intfname=xpcomintfname, paramspec=paramspec)
            self._impl.inc_indent()
            self._impl.inc_indent()

            self._impl.out(corba_preinvoke)
            self._impl.out('@retval@ = corba_obj_->@name@();',
                           name=name,
                           retval=xpcom_type.assign_retval('_retval'))
            self._impl.out(corba_postinvoke)

            self._impl.dec_indent()
            self._impl.dec_indent()
            self._impl.out(TMPL_proxy_c_operation_footer)


        if not node.readonly():
            paramspec = xpcom_type.decl_xpcom('_newval', 1, 0)
            callparams = xpcom_type.call_corba_param('_newval', 1, 0)

            corba_preinvoke = xpcom_type.corba_preinvoke('_newval', 1, 0)
            corba_postinvoke = xpcom_type.corba_postinvoke('_newval', 1, 0)

            for name in node.identifiers():
                capname = 'Set' + string.upper(name[0]) + name[1:]
                xpcomintfname = convert_scoped2flat(node.declarators()[0].scopedName()[:-1])

                self._header.out(TMPL_proxy_h_operation, name=name,
                                 capname=capname, paramspec=paramspec)

                self._impl.out(TMPL_proxy_c_operation_header,
                               name=name, capname=capname,
                               intfname=xpcomintfname, paramspec=paramspec)
                self._impl.inc_indent()
                self._impl.inc_indent()

                self._impl.out(corba_preinvoke)
                self._impl.out('corba_obj_->@name@(@paramlist@);',
                               name=name, paramlist=callparams)

                self._impl.out(corba_postinvoke)

                self._impl.dec_indent()
                self._impl.dec_indent()
                self._impl.out(TMPL_proxy_c_operation_footer)


    def visitOperation(self, node):
        name = node.identifier()
        capname = string.upper(name[0]) + name[1:]
        xpcomintfname = convert_scoped2flat(node.scopedName()[:-1])


        paramspec = []
        callparams = []
        corba_postinvoke = []
        corba_preinvoke = []
        for p in node.parameters():
            xpcom_type = xpcorba_type.get_type(p.paramType())

            paramspec.append(xpcom_type.decl_xpcom(p.identifier(),
                                                   p.is_in(), p.is_out()))
            callparams.append(xpcom_type.call_corba_param(p.identifier(),
                                                          p.is_in(),
                                                          p.is_out()))

            xpcom_corba_preinvoke = xpcom_type.corba_preinvoke(p.identifier(),
                                                               p.is_in(),
                                                               p.is_out())
            xpcom_corba_postinvoke = xpcom_type.corba_postinvoke(p.identifier(),
                                                                 p.is_in(),
                                                                 p.is_out())

            if xpcom_corba_preinvoke != '':
                corba_preinvoke.append(xpcom_corba_preinvoke)
            if xpcom_corba_postinvoke != '':
                corba_postinvoke.append(xpcom_corba_postinvoke)

        rettype = node.returnType()
        if not (isinstance(rettype, idltype.Base) and
                (rettype.kind() == idltype.tk_void)):
            retvoid = 0
            xpcom_type = xpcorba_type.get_type(node.returnType())

            paramspec.append(xpcom_type.decl_xpcom('_retval', 0, 1))

            xpcom_corba_preinvoke = xpcom_type.corba_preinvoke('_retval', 0, 1)
            xpcom_corba_postinvoke = xpcom_type.corba_postinvoke('_retval', 0, 1)

            if xpcom_corba_preinvoke != '':
                corba_preinvoke.append(xpcom_corba_preinvoke)
            if xpcom_corba_postinvoke != '':
                corba_postinvoke.append(xpcom_corba_postinvoke)
        else:
            retvoid = 1

        self._header.out(TMPL_proxy_h_operation, name=name, capname=capname,
                         paramspec=string.join(paramspec, ', '))

        self._impl.out(TMPL_proxy_c_operation_header, name=name,
                       capname=capname, intfname=xpcomintfname,
                       paramspec=string.join(paramspec, ', '))
        self._impl.inc_indent()
        self._impl.inc_indent()

        self._impl.out(string.join(corba_preinvoke, '\n'))
        if not retvoid:
            self._impl.out('@retval@ = corba_obj_->@name@(@paramlist@);',
                           name=name,
                           retval=xpcom_type.assign_retval('_retval'),
                           paramlist=string.join(callparams, ', '))
        else:
            self._impl.out('corba_obj_->@name@(@paramlist@);',
                           name=name,
                           paramlist=string.join(callparams, ', '))
        self._impl.out(string.join(corba_postinvoke, '\n'))

        self._impl.dec_indent()
        self._impl.dec_indent()
        self._impl.out(TMPL_proxy_c_operation_footer)
