#!/usr/bin/python
# Copyright (C) 2001-2002,  Christof Meerwald <cmeerw@web.de>
# http://xpcom2corba.sunsite.dk/

# XPCOM-CORBA is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.

import string

from omniidl import idlast, idlvisitor, idlutil, idltype


def convert_scoped2flat(scopedname):
    # process "#pragma flat_name"
    for pragma in idlast.findDecl(scopedname).pragmas():
        split_text = string.split(pragma.text())
        if split_text[0] == 'flat_name':
            return split_text[1]

    flat_name = 'I' + scopedname[-1]

    for i in range(len(scopedname) - 2, -1, -1):
        scope_prefix = scopedname[i] + '_'
        # process "#pragma flat_prefix"
        for pragma in idlast.findDecl(scopedname[:i + 1]).pragmas():
            split_text = string.split(pragma.text())
            if split_text[0] == 'flat_prefix':
                if (len(split_text) > 1) and (split_text[1] != '""'):
                    scope_prefix = split_text[1]
                else:
                    scope_prefix = ''
        flat_name = scope_prefix + flat_name

    return flat_name


def get_type(omniidl_type):
    unaliased_type = omniidl_type.unalias()

    if isinstance(unaliased_type, idltype.Base):
        kind = unaliased_type.kind()
        if ((kind == idltype.tk_boolean) or
            (kind == idltype.tk_long) or (kind == idltype.tk_ulong)):
            return Base_Convert(unaliased_type)
        else:
            return Base(unaliased_type)
    elif isinstance(unaliased_type, idltype.String):
        return String(unaliased_type, omniidl_type)
    elif isinstance(unaliased_type, idltype.WString):
        return WString(unaliased_type, omniidl_type)
    elif isinstance(unaliased_type, idltype.Sequence):
        return Sequence(omniidl_type, unaliased_type)
    elif isinstance(unaliased_type, idltype.Declared):
        if unaliased_type.kind() == idltype.tk_struct:
            return Struct(unaliased_type)
        elif unaliased_type.kind() == idltype.tk_objref:
            return Objref(unaliased_type)
        elif unaliased_type.kind() == idltype.tk_enum:
            return Enum(unaliased_type)
        else:
            return Declared(unaliased_type)


def temp_identifier(ident):
    return string.translate(ident, string.maketrans('->.', '___')) + '_conv'


class Type:
    def __init__(self, omniidl_type):
        pass

    def type_xpcom(self):
        return ''

    def type_corba(self):
        return ''

    def type_xpidl(self):
        return ''

    def decl_corba(self, identifier, is_in, is_out):
        return ''

    def decl_xpcom(self, identifier, is_in, is_out):
        return ''

    def decl_xpidl(self, identifier, is_in, is_out):
        if is_in and is_out:
            prefix = 'inout '
        elif is_out:
            prefix = 'out '
        elif is_in:
            prefix = 'in '
        else:
            prefix = ''

        return prefix + self.type_xpidl() + ' ' + identifier

    def call_corba_param(self, identifier, is_in, is_out):
        return ''

    def call_xpcom_param(self, identifier, is_in, is_out):
        return ''

    def assign_retval(self, identifier):
        return identifier

    def copy_xpcom2corba(self, xpcom_identifier, corba_identifier):
        return corba_identifier + ' = ' + xpcom_identifier + ';';

    def copy_corba2xpcom(self, corba_identifier, xpcom_identifier):
        return xpcom_identifier + ' = ' + corba_identifier + ';';

    def corba_preinvoke(self, identifier, is_in, is_out):
        if is_out:
            s = self.type_corba() + ' ' + temp_identifier(identifier) + ';'

            if is_in:
                s = s + ' ' + self.copy_xpcom2corba('*' + identifier,
                                                    temp_identifier(identifier))
            return s
        else:
            return ''

    def xpcom_preinvoke(self, identifier, is_in, is_out):
        if is_out:
            s = self.type_xpcom() + ' ' + temp_identifier(identifier) + ';'

            if is_in:
                s = s + ' ' + self.copy_corba2xpcom(identifier,
                                                    temp_identifier(identifier))
            return s
        else:
            return ''

    def corba_postinvoke(self, identifier, is_in, is_out):
        if is_out:
            return self.copy_corba2xpcom(temp_identifier(identifier),
                                         '*' + identifier)
        else:
            return ''

    def xpcom_postinvoke(self, identifier, is_in, is_out):
        if is_out:
            return self.copy_xpcom2corba(temp_identifier(identifier),
                                         identifier)
        else:
            return ''


class Base(Type):
    def __init__(self, omniidl_type):
        Type.__init__(self, omniidl_type)

        if omniidl_type.kind() == idltype.tk_short:
            self._type_xpcom = 'PRInt16'
            self._type_corba = 'CORBA::Short'
            self._type_xpidl = 'short'
        elif omniidl_type.kind() == idltype.tk_long:
            self._type_xpcom = 'PRInt32'
            self._type_corba = 'CORBA::Long'
            self._type_xpidl = 'long'
        elif omniidl_type.kind() == idltype.tk_ushort:
            self._type_xpcom = 'PRUint16'
            self._type_corba = 'CORBA::UShort'
            self._type_xpidl = 'unsigned short'
        elif omniidl_type.kind() == idltype.tk_ulong:
            self._type_xpcom = 'PRUint32'
            self._type_corba = 'CORBA::ULong'
            self._type_xpidl = 'unsigned long'
        elif omniidl_type.kind() == idltype.tk_float:
            self._type_xpcom = 'float'
            self._type_corba = 'CORBA::Float'
            self._type_xpidl = 'float'
        elif omniidl_type.kind() == idltype.tk_double:
            self._type_xpcom = 'double'
            self._type_corba = 'CORBA::Double'
            self._type_xpidl = 'double'
        elif omniidl_type.kind() == idltype.tk_boolean:
            self._type_xpcom = 'PRBool'
            self._type_corba = 'CORBA::Boolean'
            self._type_xpidl = 'boolean'
        elif omniidl_type.kind() == idltype.tk_char:
            self._type_xpcom = 'char'
            self._type_corba = 'CORBA::Char'
            self._type_xpidl = 'char'
        elif omniidl_type.kind() == idltype.tk_octet:
            self._type_xpcom = 'PRUint8'
            self._type_corba = 'CORBA::Octet'
            self._type_xpidl = 'octet'
        else:
            self._type_xpcom = '/* unknown */'
            self._type_corba = '/* unknown */'
            self._type_xpidl = '/* unknown */'

    def type_xpcom(self):
        return self._type_xpcom

    def type_corba(self):
        return self._type_corba

    def type_xpidl(self):
        return self._type_xpidl

    def decl_corba(self, identifier, is_in, is_out):
        if is_out:
            if is_in or identifier:
                return self._type_corba + ' &' + identifier
            else:
                return self._type_corba + ' ' + identifier
        else:
            return self._type_corba + ' ' + identifier

    def decl_xpcom(self, identifier, is_in, is_out):
        if is_out:
            return self._type_xpcom + ' *' + identifier
        else:
            return self._type_xpcom + ' ' + identifier

    def call_corba_param(self, identifier, is_in, is_out):
        if is_out:
            return '*' + identifier
        else:
            return identifier

    def call_xpcom_param(self, identifier, is_in, is_out):
        if is_out:
            return '&' + identifier
        else:
            return identifier

    def assign_retval(self, identifier):
        return '*' + identifier

    def corba_preinvoke(self, identifier, is_in, is_out):
        return ''

    def xpcom_preinvoke(self, identifier, is_in, is_out):
        return ''

    def corba_postinvoke(self, identifier, is_in, is_out):
        return ''

    def xpcom_postinvoke(self, identifier, is_in, is_out):
        return ''


class Base_Convert(Base):
    def __init__(self, omniidl_type):
        Base.__init__(self, omniidl_type)

    def call_corba_param(self, identifier, is_in, is_out):
        if is_out:
            return temp_identifier(identifier)
        else:
            return identifier

    def call_xpcom_param(self, identifier, is_in, is_out):
        if is_out:
            return '&' + temp_identifier(identifier)
        else:
            return identifier

    def assign_retval(self, identifier):
        return temp_identifier(identifier)

    def corba_preinvoke(self, identifier, is_in, is_out):
        if is_out:
            if is_in:
                return self.type_corba() + ' ' + temp_identifier(identifier) + ' = *' + identifier + ';'
            else:
                return self.type_corba() + ' ' + temp_identifier(identifier) + ';'
        else:
            return ''

    def xpcom_preinvoke(self, identifier, is_in, is_out):
        if is_out:
            if is_in:
                return self.type_xpcom() + ' ' + temp_identifier(identifier) + ' = ' + identifier + ';'
            else:
                return self.type_xpcom() + ' ' + temp_identifier(identifier) + ';'
        else:
            return ''

    def corba_postinvoke(self, identifier, is_in, is_out):
        if is_out:
            return '*' + identifier + ' = ' + temp_identifier(identifier) + ';'
        else:
            return ''

    def xpcom_postinvoke(self, identifier, is_in, is_out):
        if is_out:
            return identifier + ' = ' + temp_identifier(identifier) + ';'
        else:
            return ''


class String(Type):
    def __init__(self, unaliased_type, omniidl_type):
        Type.__init__(self, unaliased_type)
        self._utf8_encoded = 0

        t = omniidl_type
        while isinstance(t, idltype.Declared):
            if t.name()[:4] == 'utf8':
                self._utf8_encoded = 1
                break
            t = t.decl().alias().aliasType()

    def type_xpcom(self):
        if self._utf8_encoded:
            return 'nsACString'
        else:
            return 'char *'

    def type_corba(self):
        return 'char *'

    def type_xpidl(self):
        if self._utf8_encoded:
            return 'AUTF8String'
        else:
            return 'string'

    def decl_corba(self, identifier, is_in, is_out):
        if is_out:
            if is_in:
                return 'char *&' + identifier
            elif identifier:
                return 'CORBA::String_out ' + identifier
            else:
                return 'char *' + identifier
        else:
            return 'const char *' + identifier

    def decl_xpcom(self, identifier, is_in, is_out):
        if self._utf8_encoded:
            if is_out:
                return 'nsACString &' + identifier
            else:
                return 'const nsACString &' + identifier
        else:
            if is_out:
                return 'char **' + identifier
            else:
                return 'const char *' + identifier

    def call_corba_param(self, identifier, is_in, is_out):
        if is_out or self._utf8_encoded:
            return temp_identifier(identifier)
        else:
            return identifier

    def call_xpcom_param(self, identifier, is_in, is_out):
        if self._utf8_encoded:
            return temp_identifier(identifier)
        elif is_out:
            return '&' + temp_identifier(identifier)
        else:
            return identifier

    def assign_retval(self, identifier):
        return temp_identifier(identifier)

    def copy_corba2xpcom(self, corba_identifier, xpcom_identifier):
        if self._utf8_encoded:
            return '::corba2xpcom(' + corba_identifier + ', ' + xpcom_identifier + ');'
        else:
            return '*' + xpcom_identifier + ' = (char *) nsMemory::Clone(' + corba_identifier + ', sizeof (char) * (strlen(' + corba_identifier + ') + 1));'

    def copy_xpcom2corba(self, xpcom_identifier, corba_identifier):
        if self._utf8_encoded:
            return '::xpcom2corba(' + xpcom_identifier + ', ' + corba_identifier + ');'
        else:
            return corba_identifier + ' = CORBA::string_dup(' + xpcom_identifier + '); nsMemory::Free(' + xpcom_identifier + ');'

    def corba_preinvoke(self, identifier, is_in, is_out):
        if self._utf8_encoded:
            s = 'CORBA::String_var ' + temp_identifier(identifier) + ';\n'
            if is_in:
                if is_out:
                    s = s + self.copy_xpcom2corba('*' + identifier,
                                                  temp_identifier(identifier))
                else:
                    s = s + self.copy_xpcom2corba(identifier,
                                                  temp_identifier(identifier))
        elif is_out:
            if is_in:
                s = 'CORBA::String_var ' + temp_identifier(identifier) + ' = CORBA::string_dup(*' + identifier + ');'
            else:
                s = 'CORBA::String_var ' + temp_identifier(identifier) + ';'
        else:
            s = ''

        return s

    def corba_postinvoke(self, identifier, is_in, is_out):
        if is_out:
            return self.copy_corba2xpcom(temp_identifier(identifier),
                                         identifier)
        else:
            return ''

    def xpcom_preinvoke(self, identifier, is_in, is_out):
        if self._utf8_encoded:
            s = 'nsCString ' + temp_identifier(identifier) + ';\n'
            if is_in:
                s = s + self.copy_corba2xpcom(identifier,
                                              temp_identifier(identifier))
        else:
            if is_out:
                if is_in:
                    s = 'char *' + temp_identifier(identifier) + ' = (char *) nsMemory::Clone(' + identifier + ', sizeof (char) * (strlen(' + identifier + ') + 1));'
                else:
                    s = 'char *' + temp_identifier(identifier) + ';'
            else:
                s = ''

        return s


class WString(Type):
    def __init__(self, unaliased_type, omniidl_type):
        Type.__init__(self, unaliased_type)
        self._utf8_encoded = 0

        t = omniidl_type
        while isinstance(t, idltype.Declared):
            if t.name()[:4] == 'utf8':
                self._utf8_encoded = 1
                break
            t = t.decl().alias().aliasType()

    def type_xpcom(self):
        if self._utf8_encoded:
            return 'nsACString'
        else:
            return 'PRUnichar *'

    def type_corba(self):
        return 'CORBA::WChar *'

    def type_xpidl(self):
        if self._utf8_encoded:
            return 'AUTF8String'
        else:
            return 'wstring'

    def decl_corba(self, identifier, is_in, is_out):
        if is_out:
            if is_in:
                return 'CORBA::WChar *&' + identifier
            elif identifier:
                return 'CORBA::WString_out ' + identifier
            else:
                return 'CORBA::WChar *' + identifier
        else:
            return 'const CORBA::WChar *' + identifier

    def decl_xpcom(self, identifier, is_in, is_out):
        if self._utf8_encoded:
            if is_out:
                return 'nsACString &' + identifier
            else:
                return 'const nsACString &' + identifier
        else:
            if is_out:
                return 'PRUnichar **' + identifier
            else:
                return 'const PRUnichar *' + identifier

    def call_corba_param(self, identifier, is_in, is_out):
        return temp_identifier(identifier)

    def call_xpcom_param(self, identifier, is_in, is_out):
        if is_out and not self._utf8_encoded:
            return '&' + temp_identifier(identifier)
        else:
            return temp_identifier(identifier)

    def assign_retval(self, identifier):
        return temp_identifier(identifier)

    def copy_corba2xpcom(self, corba_identifier, xpcom_identifier):
        return '::corba2xpcom(' + corba_identifier + ', ' + xpcom_identifier + ');'

    def copy_xpcom2corba(self, xpcom_identifier, corba_identifier):
        return '::xpcom2corba(' + xpcom_identifier + ', ' + corba_identifier + ');'

    def corba_preinvoke(self, identifier, is_in, is_out):
        s = 'CORBA::WString_var ' + temp_identifier(identifier) + ';\n'
        if is_in:
            if is_out:
                s = s + self.copy_xpcom2corba('*' + identifier,
                                              temp_identifier(identifier))
            else:
                s = s + self.copy_xpcom2corba(identifier,
                                              temp_identifier(identifier))
        return s

    def corba_postinvoke(self, identifier, is_in, is_out):
        if is_out:
            return self.copy_corba2xpcom(temp_identifier(identifier),
                                         identifier)
        else:
            return ''

    def xpcom_preinvoke(self, identifier, is_in, is_out):
        if self._utf8_encoded:
            s = 'nsCString ' + temp_identifier(identifier) + ';\n'
            if is_in:
                s = s + self.copy_corba2xpcom(identifier,
                                              temp_identifier(identifier))
        else:
            s = 'PRUnichar *' + temp_identifier(identifier) + ';\n'
            if is_in:
                s = s + self.copy_corba2xpcom(identifier,
                                              '&' + temp_identifier(identifier))

        return s


class Sequence(Type):
    def __init__(self, omniidl_type, unaliased_type):
        Type.__init__(self, omniidl_type)

        # TODO: XPCOM doesn't support sequence of sequence yet
        self._child = get_type(unaliased_type.seqType())
        self._corbaname = idlutil.ccolonName(omniidl_type.scopedName())

    def type_corba(self):
        return self._corbaname + ' *'

    # TODO
    #def type_xpcom(self):
    
    def type_xpidl(self):
        return ''

    def decl_corba(self, identifier, is_in, is_out):
        if is_out:
            if is_in:
                return self._corbaname + ' &' + identifier
            elif identifier:
                return self._corbaname + '_out ' + identifier
            else:
                return self._corbaname + ' *'
        else:
            return 'const ' + self._corbaname + ' &' + identifier

    def decl_xpcom(self, identifier, is_in, is_out):
        if is_out:
            s = self._child.decl_xpcom('*' + identifier, is_in, is_out) + ', PRUint32 *_sizeof_' + identifier
        else:
            # NOTE: change to 'const *' as soon as xpidl gets it right
            s = self._child.decl_xpcom('*' + identifier, is_in, is_out) + ', PRUint32 _sizeof_' + identifier
        return s

    def decl_xpidl(self, identifier, is_in, is_out):
        if is_in and is_out:
            prefix = 'inout '
        elif is_out:
            prefix = 'out '
        elif is_in:
            prefix = 'in '
        else:
            prefix = ''

        return '[array, size_is(sizeof_' + identifier + ')] ' + prefix + self._child.type_xpidl() + ' ' + identifier + ', ' + prefix + 'PRUint32 sizeof_' + identifier

    def call_corba_param(self, identifier, is_in, is_out):
        return temp_identifier(identifier)

    def call_xpcom_param(self, identifier, is_in, is_out):
        if is_out:
            return '&' + temp_identifier(identifier) + ', &_sizeof_' + temp_identifier(identifier)
        elif isinstance(self._child, String):
            # workaround for non-const XPCOM array parameters
            return 'const_cast<const char **>(' + temp_identifier(identifier) + '), _sizeof_' + temp_identifier(identifier)
        elif isinstance(self._child, WString):
            # workaround for non-const XPCOM array parameters
            return 'const_cast<const PRUnichar **>(' + temp_identifier(identifier) + '), _sizeof_' + temp_identifier(identifier)
        else:
            return temp_identifier(identifier) + ', _sizeof_' + temp_identifier(identifier)

    def assign_retval(self, identifier):
        return temp_identifier(identifier)

    def copy_xpcom2corba(self, xpcom_identifier, corba_identifier):
        if xpcom_identifier[0] == '*':
            return '::xpcom2corba(' + xpcom_identifier + ', *_sizeof_' + xpcom_identifier[1:] + ', ' + corba_identifier + ');'
        else:
            return '::xpcom2corba(' + xpcom_identifier + ', _sizeof_' + xpcom_identifier + ', ' + corba_identifier + ');'

    def copy_corba2xpcom(self, corba_identifier, xpcom_identifier):
        if (xpcom_identifier[0] == '*') or (xpcom_identifier[0] == '&'):
            return '::corba2xpcom(' + corba_identifier + ', ' + xpcom_identifier + ', ' + xpcom_identifier[0] + '_sizeof_' + xpcom_identifier[1:] + ');'
        else:
            return '::corba2xpcom(' + corba_identifier + ', ' + xpcom_identifier + ', _sizeof_' + xpcom_identifier + ');'

    def xpcom_preinvoke(self, identifier, is_in, is_out):
        s = self._child.type_xpcom() + ' *' + temp_identifier(identifier) + '; PRUint32 _sizeof_' + temp_identifier(identifier) + ';'
        if is_in:
            s = s + '\n' + self.copy_corba2xpcom(identifier, '&' + temp_identifier(identifier))

        return s

    def corba_preinvoke(self, identifier, is_in, is_out):
        s = self._corbaname + '_var ' + temp_identifier(identifier)
        if is_in:
            if is_out:
                deref = '*'
            else:
                deref = ''
            s = s + ' = new ' + self._corbaname + '();\n' + self.copy_xpcom2corba(deref + identifier, temp_identifier(identifier))
        else:
            s = s + ';'

        return s

    def xpcom_postinvoke(self, identifier, is_in, is_out):
        if is_out:
            if is_in:
                return self.copy_xpcom2corba(temp_identifier(identifier),
                                             identifier)
            else:
                s = identifier + ' = new ' + self._corbaname + '(); '
                s = s + self.copy_xpcom2corba(temp_identifier(identifier),
                                              '*' + identifier)
                return s
        else:
            return ''

    def corba_postinvoke(self, identifier, is_in, is_out):
        if is_out:
            return self.copy_corba2xpcom(temp_identifier(identifier),
                                         identifier)
        else:
            return ''


class Declared(Type):
    def __init__(self, omniidl_type):
        Type.__init__(self, omniidl_type)

        self._intfname = convert_scoped2flat(omniidl_type.scopedName())
        self._corbaname = idlutil.ccolonName(omniidl_type.scopedName())


    def type_corba(self):
        return self._corbaname + '_ptr'

    def type_xpcom(self):
        return self._intfname + ' *'

    def type_xpidl(self):
        if self._intfname == 'CORBA_IObject':
            return 'nsISupports'
        else:
            return self._intfname

    def decl_corba(self, identifier, is_in, is_out):
        if is_out:
            if is_in:
                return self._corbaname + '_ptr &' + identifier
            elif identifier:
                return self._corbaname + '_out ' + identifier
            else:
                return self._corbaname + '_ptr ' + identifier
        else:
            return self._corbaname + '_ptr ' + identifier

    def decl_xpcom(self, identifier, is_in, is_out):
        if is_out:
            return self._intfname + ' **' + identifier
        else:
            return self._intfname + ' *' + identifier

    def call_corba_param(self, identifier, is_in, is_out):
        return temp_identifier(identifier)

    def call_xpcom_param(self, identifier, is_in, is_out):
        if is_out:
            return '&' + temp_identifier(identifier)
        else:
            return temp_identifier(identifier)

    def assign_retval(self, identifier):
        return temp_identifier(identifier)

    def copy_xpcom2corba(self, xpcom_identifier, corba_identifier):
        return '{' + self._corbaname + '* tmptype = (' + self._corbaname + "*)NULL;\n" + \
               self._corbaname + "_ptr tmpptype;\n" + \
               '::xpcom2corba(' + xpcom_identifier + ", tmptype, &tmpptype);\n" + \
               corba_identifier + " = tmpptype;\n}"

    def copy_corba2xpcom(self, corba_identifier, xpcom_identifier):
        return xpcom_identifier + ' = new ' + self._intfname + '__XPCOM_Proxy(' + corba_identifier + '); NS_ADDREF(' + xpcom_identifier + ');'

    def corba_preinvoke(self, identifier, is_in, is_out):
        if is_in:
            if is_out:
                deref = '*'
            else:
                deref = ''
            return self._corbaname + '* ' + temp_identifier(identifier) +\
                   '_type = (' + self._corbaname + "*)NULL;\n" +\
                   self._corbaname + '_var ' + temp_identifier(identifier) + ";\n" +\
                   '{' + self._corbaname + '_ptr ' + temp_identifier(identifier) + "_ptr;\n" +\
                   '::xpcom2corba(' + deref + identifier + ', ' +\
                   temp_identifier(identifier) + '_type, &' +\
                   temp_identifier(identifier) + "_ptr);\n" +\
                   temp_identifier(identifier) + ' = ' +\
                   temp_identifier(identifier) + "_ptr;\n}\n";
        else:
            return self._corbaname + '_var ' + temp_identifier(identifier) + ';'

    def xpcom_preinvoke(self, identifier, is_in, is_out):
        if is_in:
            return self._intfname + ' *' + temp_identifier(identifier) + ' = new ' + self._intfname + '__XPCOM_Proxy(' + identifier + '); NS_ADDREF(' + temp_identifier(identifier) + ');'
        else:
            return self._intfname + ' *' + temp_identifier(identifier) + ';'

    def corba_postinvoke(self, identifier, is_in, is_out):
        if is_out:
            return self.copy_corba2xpcom(temp_identifier(identifier),
                                         '*' + identifier)
        else:
            return ''

    def xpcom_postinvoke(self, identifier, is_in, is_out):
        if is_out:
            return self.copy_xpcom2corba(temp_identifier(identifier),
                                         identifier)
        else:
            return ''


class Struct(Declared):
    def __init__(self, omniidl_type):
        Declared.__init__(self, omniidl_type)

class Objref(Declared):
    def __init__(self, omniidl_type):
        Declared.__init__(self, omniidl_type)

    def copy_corba2xpcom(self, corba_identifier, xpcom_identifier):
        s = 'if (!CORBA::is_nil(' + corba_identifier + '))\n{\n'
        if self._intfname == 'CORBA_IObject':
            s = s + '  XPCOM::IObject_var xpcom_iobject = XPCOM::IObject::_narrow(' + corba_identifier + ');\n'
            s = s + '  if (!CORBA::is_nil(xpcom_iobject))\n  {\n'
            s = s + '    ' + xpcom_identifier + ' = new XPCOM_IObject__XPCOM_Proxy(xpcom_iobject); NS_ADDREF(' + xpcom_identifier + ');\n'
            s = s + '  }\n  else\n  {\n'
            s = s + '    ' + xpcom_identifier + ' = new CORBA_IObject__XPCOM_Proxy(' + corba_identifier + '); NS_ADDREF(' + xpcom_identifier + ');\n'
            s = s + '  }\n'
        else:
            s = s + '  ' + xpcom_identifier + ' = new ' + self._intfname + '__XPCOM_Proxy(' + corba_identifier + '); NS_ADDREF(' + xpcom_identifier + ');\n'
        s = s + '}'
        return s

    def xpcom_preinvoke(self, identifier, is_in, is_out):
        s = self._intfname + ' *' + temp_identifier(identifier)
        if is_in:
            s = s + ' = NULL;\n'
            s = s + self.copy_corba2xpcom(identifier,
                                          temp_identifier(identifier))
            return s
        else:
            return s + ';'

    def corba_postinvoke(self, identifier, is_in, is_out):
        if is_out:
            s = 'if (!CORBA::is_nil(' + temp_identifier(identifier) + '))\n{\n'
            if self._intfname == 'CORBA_IObject':
                s = s + '  XPCOM::IObject_var xpcom_iobject = XPCOM::IObject::_narrow(' + temp_identifier(identifier) + ');\n'
                s = s + '  if (!CORBA::is_nil(xpcom_iobject))\n  {\n'
                s = s + '    *' + identifier + ' = new XPCOM_IObject__XPCOM_Proxy(xpcom_iobject); NS_ADDREF(*' + identifier + ');\n'
                s = s + '  }\n  else\n  {\n'
                s = s + '    *' + identifier + ' = new CORBA_IObject__XPCOM_Proxy(' + temp_identifier(identifier) + '); NS_ADDREF(*' + identifier + ');\n'
                s = s + '  }\n'
            else:
                s = s + '  *' + identifier + ' = new ' + self._intfname + '__XPCOM_Proxy(' + temp_identifier(identifier) + '); NS_ADDREF(*' + identifier + ');\n'
            s = s + '}\nelse\n{\n  *' + identifier + ' = NULL;\n}'
            return s
        else:
            return ''


class Enum(Declared):
    def __init__(self, omniidl_type):
        Declared.__init__(self, omniidl_type)

    def type_corba(self):
        return self._corbaname

    def type_xpcom(self):
        return 'PRUint32'

    def type_xpidl(self):
        return 'unsigned long'

    def decl_corba(self, identifier, is_in, is_out):
        if is_out:
            if is_in:
                return self._corbaname + ' &' + identifier
            elif identifier:
                return self._corbaname + ' &' + identifier
            else:
                return self._corbaname + ' ' + identifier
        else:
            return self._corbaname + ' ' + identifier

    def decl_xpcom(self, identifier, is_in, is_out):
        if is_out:
            return self.type_xpcom() + ' *' + identifier
        else:
            return self.type_xpcom() + ' ' + identifier

    def call_corba_param(self, identifier, is_in, is_out):
        if is_out:
            return temp_identifier(identifier)
        else:
            return self.type_corba() + '(' + identifier + ')'

    def call_xpcom_param(self, identifier, is_in, is_out):
        if is_out:
            return '&' + temp_identifier(identifier)
        else:
            return identifier

    def assign_retval(self, identifier):
        return '*' + identifier

    def copy_corba2xpcom(self, corba_identifier, xpcom_identifier):
        return xpcom_identifier + ' = ' + corba_identifier + ';'

    def copy_xpcom2corba(self, xpcom_identifier, corba_identifier):
        return corba_identifier + ' = ' + self._corbaname + '(' + xpcom_identifier + ');'

    def corba_preinvoke(self, identifier, is_in, is_out):
        if is_out:
            s = self.type_corba() + ' ' + temp_identifier(identifier)

            if is_in:
                return s + ' = ' + self._corbaname + '(*' + identifier + ');'
            else:
                return s + ';'
        else:
            return ''

    def xpcom_preinvoke(self, identifier, is_in, is_out):
        if is_out:
            s = self.type_xpcom() + ' ' + temp_identifier(identifier)

            if is_in:
                s = s + ' = ' + identifier
            return s + ';'
        else:
            return ''

    def corba_postinvoke(self, identifier, is_in, is_out):
        return Type.corba_postinvoke(self, identifier, is_in, is_out)

    def xpcom_postinvoke(self, identifier, is_in, is_out):
        return Type.xpcom_postinvoke(self, identifier, is_in, is_out)
